/*
 * "$Id$"
 *
 *   OpenGL bitmap demo.
 *
 * Contents:
 *
 *   WinMain()             - Main entry...
 *   DisplayErrorMessage() - Display an error message dialog.
 *   MakePalette()         - Make a color palette for RGB colors if necessary.
 *   BitmapProc()          - Handle window events in the viewing window.
 *   RepaintWindow()       - Redraw the client area with our bitmap...
 *
 * Revision History:
 *
 *   $Log$
 */

/*
 * Include necessary headers.
 */

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "oglbits.h"
#include <stdarg.h>


/*
 * Globals...
 */

HWND		BitmapWindow;		/* Bitmap window */
HPALETTE	BitmapPalette;		/* Color palette (if necessary) */
HDC		BitmapDC;		/* Drawing context */
HGLRC		BitmapRC;		/* OpenGL rendering context */


/*
 * Local functions...
 */

void			DisplayErrorMessage(char *, ...);
void			MakePalette(int);
LRESULT CALLBACK	BitmapProc(HWND, UINT, WPARAM, LPARAM);
void			RepaintWindow(RECT *);


/*
 * 'WinMain()' - Main entry...
 */

int APIENTRY
WinMain(HINSTANCE hInst,		/* I - Current process instance */
	HINSTANCE hPrevInstance,	/* I - Parent process instance */
	LPSTR     lpCmdLine,		/* I - Command-line arguments */
	int       nCmdShow)		/* I - Show window at startup? */
{
  MSG		msg;			/* Window UI event */
  WNDCLASS	wc;                     /* Window class */


 /*
  * Register main window...
  */

  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)BitmapProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInst;
  wc.hIcon         = NULL;
  wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = 0;         
  wc.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
  wc.lpszClassName = "OpenGL Bitmap Demo";

  if (RegisterClass(&wc) == 0)
  {
    DisplayErrorMessage("Unable to register window class!");
    return (FALSE);
  };

 /*
  * Then create it...
  */

  BitmapWindow = CreateWindow("OpenGL Bitmap Demo", "OpenGL Bitmap Demo",
                              WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
  			      32, 32, 256, 256,
			      NULL, NULL, hInst, NULL);

  if (BitmapWindow == NULL)
  {
    DisplayErrorMessage("Unable to create window!");
    return (FALSE);
  };

  ShowWindow(BitmapWindow, nCmdShow);
  UpdateWindow(BitmapWindow);

 /*
  * Loop on events until the user quits this application...
  */

  while (GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  };

  return (msg.wParam);
}


/*
 * 'DisplayErrorMessage()' - Display an error message dialog.
 */

void
DisplayErrorMessage(char *format,	/* I - printf() style format string */
                    ...)		/* I - Other arguments as necessary */
{
  va_list	ap;			/* Argument pointer */
  char		s[1024];		/* Output string */


  if (format == NULL)
    return;

  va_start(ap, format);
  vsprintf(s, format, ap);
  va_end(ap);

  MessageBeep(MB_ICONEXCLAMATION);
  MessageBox(NULL, s, "Error", MB_OK | MB_ICONEXCLAMATION);
}


/*
 * 'BitmapProc()' - Handle window events in the viewing window.
 */

LRESULT CALLBACK
BitmapProc(HWND   hWnd,		/* I - Window triggering this event */
	 UINT   uMsg,		/* I - Message type */
	 WPARAM wParam,		/* I - 'word' parameter value */
	 LPARAM lParam)		/* I - 'long' parameter value */
{
  int			pf;	/* Pixel format ID */
  PIXELFORMATDESCRIPTOR pfd;	/* Pixel format information */
  PAINTSTRUCT		ps;
  RECT			rect;


  switch (uMsg)
  {
    case WM_CREATE :
       /*
        * 'Create' message.  Get device and rendering contexts, and
	* setup the client area for OpenGL drawing...
	*/

	BitmapDC = GetDC(hWnd);              

        pfd.nSize        = sizeof(pfd);
        pfd.nVersion     = 1;
        pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
						/* Do OpenGL drawing */
        pfd.dwLayerMask  = PFD_MAIN_PLANE;	/* Main drawing plane */
        pfd.iPixelType   = PFD_TYPE_COLORINDEX;	/* Color index buffer */
        pfd.cColorBits   = 0;			/* Best color buffer please */
        pfd.cDepthBits   = 0;			/* No depth buffer */
        pfd.cStencilBits = 0;			/* No stencil buffer */
        pfd.cAccumBits   = 0;			/* No accumulation buffer */

        pf = ChoosePixelFormat(BitmapDC, &pfd);
	if (pf == 0)
          DisplayErrorMessage("oglbits was unable to choose a suitable pixel format!");
        else if (!SetPixelFormat(BitmapDC, pf, &pfd))
          DisplayErrorMessage("oglbits was unable to set the pixel format!");

        BitmapRC = wglCreateContext(BitmapDC);
        wglMakeCurrent(BitmapDC, BitmapRC);
	break;

    case WM_SIZE :
    case WM_PAINT :
       /*
        * Repaint the client area with our bitmap...
	*/

	BeginPaint(hWnd, &ps);

        GetClientRect(hWnd, &rect);
	RepaintWindow(&rect);

	EndPaint(hWnd, &ps);
	break;

    case WM_COMMAND :
       /*
        * Handle menu selections...
	*/

	switch (LOWORD(wParam))
	{
	  case IDM_FILE_EXIT :
	      DestroyWindow(BitmapWindow);
	      break;
	};
        break;

    case WM_QUIT :
    case WM_CLOSE :
       /*
        * Destroy the viewing window and exit...
        */
        
        DestroyWindow(BitmapWindow);
	exit(0);
        break;

    case WM_DESTROY :
       /*
        * Release and free the device context and rendering context
	*/
  	if (BitmapRC)
    	  wglDeleteContext(BitmapRC);

    	if (BitmapDC)
    	  ReleaseDC(BitmapWindow, BitmapDC);

	if (BitmapPalette)
	  DeleteObject(BitmapPalette);

        PostQuitMessage(0);
        break;

    default :
       /*
        * Pass all other messages through the default window
        * procedure...
        */

        return (DefWindowProc(hWnd, uMsg, wParam, lParam));
  };

  return (FALSE);
}


/*
 * 'RepaintWindow()' - Redraw the client area with our bitmap...
 */

void
RepaintWindow(RECT *rect)	/* I - Client area rectangle */
{
  int		i;		/* Looping var */
  static GLubyte smiley[] =	/* 16x16 smiley face */
  {
    0x03, 0xc0, 0, 0, /*       ****       */
    0x0f, 0xf0, 0, 0, /*     ********     */
    0x1e, 0x78, 0, 0, /*    ****  ****    */
    0x39, 0x9c, 0, 0, /*   ***  **  ***   */
    0x77, 0xee, 0, 0, /*  *** ****** ***  */
    0x6f, 0xf6, 0, 0, /*  ** ******** **  */
    0xff, 0xff, 0, 0, /* **************** */
    0xff, 0xff, 0, 0, /* **************** */
    0xff, 0xff, 0, 0, /* **************** */
    0xff, 0xff, 0, 0, /* **************** */
    0x73, 0xce, 0, 0, /*  ***  ****  ***  */
    0x73, 0xce, 0, 0, /*  ***  ****  ***  */
    0x3f, 0xfc, 0, 0, /*   ************   */
    0x1f, 0xf8, 0, 0, /*    **********    */
    0x0f, 0xf0, 0, 0, /*     ********     */
    0x03, 0xc0, 0, 0  /*       ****       */
  };


  glViewport(0, 0, rect->right, rect->bottom);

  glClearIndex(0.0);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, rect->right - 1.0, 0.0, rect->bottom - 1.0, -1.0, 1.0);

 /*
  * This bitmap is aligned to 4-byte boundaries...
  */

  glPixelTransferi(GL_UNPACK_ALIGNMENT, 4);
  glPixelTransferi(GL_INDEX_OFFSET, 1);

  for (i = 0; i < 100; i ++)
  { 
    glRasterPos2i(rand() % rect->right, rand() % rect->bottom);
    glDrawPixels(16, 16, GL_COLOR_INDEX, GL_BITMAP, smiley);
  };

  glFinish();
}


/*
 * End of "$Id$".
 */
