/*
 * "$Id: oglfont.c,v 1.1 1995/12/31 18:10:10 mike Exp mike $"
 *
 *   OpenGL raster font demo.
 *
 * Contents:
 *
 *   WinMain()             - Main entry...
 *   DisplayErrorMessage() - Display an error message dialog.
 *   MakePalette()         - Make a color palette for RGB colors if necessary.
 *   FontProc()            - Handle window events in the viewing window.
 *   RepaintWindow()       - Redraw the client area with our bitmap...
 *
 * Revision History:
 *
 *   $Log: oglfont.c,v $
 *   Revision 1.1  1995/12/31  18:10:10  mike
 *   Initial revision
 */

/*
 * Include necessary headers.
 */

#include "font.h"
#include "oglfont.h"
#include <stdarg.h>


/*
 * Globals...
 */

HWND		FontWindow;		/* Font window */
HPALETTE	FontPalette;		/* Color palette (if necessary) */
HDC		FontDC;			/* Drawing context */
HGLRC		FontRC;			/* OpenGL rendering context */

GLuint		SmallFont,		/* Small raster font */
		MediumFont,		/* Medium raster font */
		LargeFont;		/* Large raster font */


/*
 * Local functions...
 */

void			DisplayErrorMessage(char *, ...);
void			MakePalette(int);
LRESULT CALLBACK	FontProc(HWND, UINT, WPARAM, LPARAM);
void			RepaintWindow(RECT *);


/*
 * 'WinMain()' - Main entry...
 */

int APIENTRY
WinMain(HINSTANCE hInst,		/* I - Current process instance */
	HINSTANCE hPrevInstance,	/* I - Parent process instance */
	LPSTR     lpCmdLine,		/* I - Command-line arguments */
	int       nCmdShow)		/* I - Show window at startup? */
{
  MSG		msg;			/* Window UI event */
  WNDCLASS	wc;                     /* Window class */


 /*
  * Register main window...
  */

  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)FontProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInst;
  wc.hIcon         = NULL;
  wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = 0;         
  wc.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
  wc.lpszClassName = "OpenGL Font Demo";

  if (RegisterClass(&wc) == 0)
  {
    DisplayErrorMessage("Unable to register window class!");
    return (FALSE);
  };

 /*
  * Then create it...
  */

  FontWindow = CreateWindow("OpenGL Font Demo", "OpenGL Font Demo",
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
  			    32, 32, 256, 256,
			    NULL, NULL, hInst, NULL);

  if (FontWindow == NULL)
  {
    DisplayErrorMessage("Unable to create window!");
    return (FALSE);
  };

  ShowWindow(FontWindow, nCmdShow);
  UpdateWindow(FontWindow);

 /*
  * Loop on events until the user quits this application...
  */

  while (GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  };

  return (msg.wParam);
}


/*
 * 'DisplayErrorMessage()' - Display an error message dialog.
 */

void
DisplayErrorMessage(char *format,	/* I - printf() style format string */
                    ...)		/* I - Other arguments as necessary */
{
  va_list	ap;			/* Argument pointer */
  char		s[1024];		/* Output string */


  if (format == NULL)
    return;

  va_start(ap, format);
  vsprintf(s, format, ap);
  va_end(ap);

  MessageBeep(MB_ICONEXCLAMATION);
  MessageBox(NULL, s, "Error", MB_OK | MB_ICONEXCLAMATION);
}


/*
 * 'MakePalette()' - Make a color palette for RGB colors if necessary.
 */

void
MakePalette(int pf)			/* I - Pixel format ID */
{
  PIXELFORMATDESCRIPTOR pfd;		/* Pixel format information */
  LOGPALETTE		*pPal;		/* Pointer to logical palette */
  int			nColors;	/* Number of entries in palette */
  int			i,
			rmax,
			gmax,
			bmax;


 /*
  * Find out if we need to define a color palette...
  */

  DescribePixelFormat(FontDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

  if (!(pfd.dwFlags & PFD_NEED_PALETTE))
  {
    FontPalette = NULL;
    return;
  };

 /*
  * Allocate memory for a color palette...
  */

  nColors = 1 << pfd.cColorBits;	

  pPal = (LOGPALETTE *)malloc(sizeof(LOGPALETTE) +
                              nColors * sizeof(PALETTEENTRY));

  pPal->palVersion    = 0x300;
  pPal->palNumEntries = nColors;

 /*
  * Get the maximum values for red, green, and blue.  Then build 'nColors'
  * colors...
  */

  rmax = (1 << pfd.cRedBits) - 1;
  gmax = (1 << pfd.cGreenBits) - 1;
  bmax = (1 << pfd.cBlueBits) - 1;

  for (i = 0; i < nColors; i ++)
  {
    pPal->palPalEntry[i].peRed   = 255 * ((i >> pfd.cRedShift) & rmax) / rmax;
    pPal->palPalEntry[i].peGreen = 255 * ((i >> pfd.cGreenShift) & gmax) / gmax;
    pPal->palPalEntry[i].peBlue  = 255 * ((i >> pfd.cBlueShift) & bmax) / bmax;

    pPal->palPalEntry[i].peFlags = 0;
  };
		
 /*
  * Create, select, and realize the palette...
  */

  FontPalette = CreatePalette(pPal);
  SelectPalette(FontDC, FontPalette, FALSE);
  RealizePalette(FontDC);

  free(pPal);
}


/*
 * 'FontProc()' - Handle window events in the viewing window.
 */

LRESULT CALLBACK
FontProc(HWND   hWnd,		/* I - Window triggering this event */
	 UINT   uMsg,		/* I - Message type */
	 WPARAM wParam,		/* I - 'word' parameter value */
	 LPARAM lParam)		/* I - 'long' parameter value */
{
  int			pf;	/* Pixel format ID */
  PIXELFORMATDESCRIPTOR pfd;	/* Pixel format information */
  PAINTSTRUCT		ps;
  RECT			rect;


  switch (uMsg)
  {
    case WM_CREATE :
       /*
        * 'Create' message.  Get device and rendering contexts, and
	* setup the client area for OpenGL drawing...
	*/

	FontDC = GetDC(hWnd);              

        pfd.nSize        = sizeof(pfd);
        pfd.nVersion     = 1;
        pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
						/* Do OpenGL drawing */
        pfd.dwLayerMask  = PFD_MAIN_PLANE;	/* Main drawing plane */
        pfd.iPixelType   = PFD_TYPE_RGBA;	/* RGB color buffer */
        pfd.cColorBits   = 0;			/* Best color buffer please */
        pfd.cDepthBits   = 0;			/* No depth buffer */
        pfd.cStencilBits = 0;			/* No stencil buffer */
        pfd.cAccumBits   = 0;			/* No accumulation buffer */

        pf = ChoosePixelFormat(FontDC, &pfd);
	if (pf == 0)
          DisplayErrorMessage("oglfont was unable to choose a suitable pixel format!");
        else if (!SetPixelFormat(FontDC, pf, &pfd))
          DisplayErrorMessage("oglfont was unable to set the pixel format!");

        MakePalette(pf);

        FontRC = wglCreateContext(FontDC);
        wglMakeCurrent(FontDC, FontRC);

        SmallFont  = FontCreateBitmaps(FontDC, "Courier", 12,  FW_BOLD, 0);
        MediumFont = FontCreateBitmaps(FontDC, "Courier", 24, FW_BOLD, 0);
        LargeFont  = FontCreateBitmaps(FontDC, "Courier", 36, FW_BOLD, 0);
	break;

    case WM_SIZE :
    case WM_PAINT :
       /*
        * Repaint the client area with our bitmap...
	*/

	BeginPaint(hWnd, &ps);

        GetClientRect(hWnd, &rect);
	RepaintWindow(&rect);

	EndPaint(hWnd, &ps);
	break;

    case WM_COMMAND :
       /*
        * Handle menu selections...
	*/

	switch (LOWORD(wParam))
	{
	  case IDM_FILE_EXIT :
	      DestroyWindow(FontWindow);
	      break;
	};
        break;

    case WM_QUIT :
    case WM_CLOSE :
       /*
        * Destroy the viewing window and exit...
        */
        
        DestroyWindow(FontWindow);
	exit(0);
        break;

    case WM_DESTROY :
       /*
        * Release and free the device context and rendering context
	*/
  	if (FontRC)
    	  wglDeleteContext(FontRC);

    	if (FontDC)
    	  ReleaseDC(FontWindow, FontDC);

	if (FontPalette)
	  DeleteObject(FontPalette);

        PostQuitMessage(0);
        break;

    case WM_QUERYNEWPALETTE :
       /*
        * Realize the color palette if necessary...
	*/

        if (FontPalette)
	{
          SelectPalette(FontDC, FontPalette, FALSE);
          RealizePalette(FontDC);

          InvalidateRect(hWnd, NULL, FALSE);
	  return (TRUE);
	};
        break;
	
    case WM_PALETTECHANGED:
       /*
        * Reselect our color palette if necessary...
	*/

	if (FontPalette && (HWND)wParam != hWnd)
	{
          SelectPalette(FontDC, FontPalette, FALSE);
          RealizePalette(FontDC);

          UpdateColors(FontDC);
	};
	break;

    default :
       /*
        * Pass all other messages through the default window
        * procedure...
        */

        return (DefWindowProc(hWnd, uMsg, wParam, lParam));
  };

  return (FALSE);
}


/*
 * 'RepaintWindow()' - Redraw the client area with our bitmap...
 */

void
RepaintWindow(RECT *rect)	/* I - Client area rectangle */
{
 /*
  * Reset the viewport and clear the window to white...
  */

  glViewport(0, 0, rect->right, rect->bottom);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, rect->right - 1.0, 0.0, rect->bottom - 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

  glColor3f(1.0, 0.0, 0.0);
  glRasterPos2i(16, 96);
  FontPuts(SmallFont, "Small Font");

  glColor3f(0.0, 1.0, 0.0);
  glRasterPos2i(16, 64);
  FontPuts(MediumFont, "Medium Font");

  glColor3f(0.1, 0.1, 1.0);
  glRasterPos2i(16, 16);
  FontPuts(LargeFont, "Large Font");

  glFinish();
}


/*
 * End of "$Id: oglfont.c,v 1.1 1995/12/31 18:10:10 mike Exp mike $".
 */
