/*
 * "$Id: oglview.c,v 1.1 1995/12/31 07:27:56 mike Exp mike $"
 *
 *   OpenGL bitmap viewing program.
 *
 * Contents:
 *
 *   WinMain()             - Main entry...
 *   DisplayErrorMessage() - Display an error message dialog.
 *   MakePalette()         - Make a color palette for RGB colors if
 *                           necessary.
 *   ViewProc()            - Handle window events in the viewing window.
 *   OpenBitmapFile()      - Load a bitmap file from disk.
 *   SaveBitmapFile()      - Save the currently displayed bitmap to disk.
 *   PrintBitmap()         - Print the currently displayed bitmap.
 *   RepaintWindow()       - Redraw the client area with our bitmap...
 *
 * Revision History:
 *
 *   $Log: oglview.c,v $
 *   Revision 1.1  1995/12/31  07:27:56  mike
 *   Initial revision
 */

/*
 * Include necessary headers.
 */

#include "bitmap.h"
#include "oglview.h"
#include <stdarg.h>


/*
 * Globals...
 */

HWND		ViewWindow;		/* View window */
HPALETTE	ViewPalette;		/* Color palette (if necessary) */
HDC		ViewDC;			/* Drawing context */
HGLRC		ViewRC;			/* OpenGL rendering context */

BITMAPINFO	*BitmapInfo = NULL;	/* Current bitmap information */
void		*BitmapBits = NULL;	/* Current bitmap pixel bits */


/*
 * Local functions...
 */

void			DisplayErrorMessage(char *, ...);
void			MakePalette(int);
LRESULT CALLBACK	ViewProc(HWND, UINT, WPARAM, LPARAM);
void			OpenBitmapFile(char *);
void			SaveBitmapFile(void);
void			PrintBitmap(void);
void			RepaintWindow(RECT *);


/*
 * 'WinMain()' - Main entry...
 */

int APIENTRY
WinMain(HINSTANCE hInst,		/* I - Current process instance */
	HINSTANCE hPrevInstance,	/* I - Parent process instance */
	LPSTR     lpCmdLine,		/* I - Command-line arguments */
	int       nCmdShow)		/* I - Show window at startup? */
{
  MSG		msg;			/* Window UI event */
  WNDCLASS	wc;                     /* Window class */


 /*
  * Register main window...
  */

  wc.style         = 0;
  wc.lpfnWndProc   = (WNDPROC)ViewProc;
  wc.cbClsExtra    = 0;
  wc.cbWndExtra    = 0;
  wc.hInstance     = hInst;
  wc.hIcon         = NULL;
  wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = 0;         
  wc.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
  wc.lpszClassName = "OpenGL Bitmap Viewer";

  if (RegisterClass(&wc) == 0)
  {
    DisplayErrorMessage("Unable to register window class!");
    return (FALSE);
  };

 /*
  * Then create it...
  */

  ViewWindow = CreateWindow("OpenGL Bitmap Viewer", "OpenGL Bitmap Viewer",
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
  			    32, 32, 256, 256,
			    NULL, NULL, hInst, NULL);

  if (ViewWindow == NULL)
  {
    DisplayErrorMessage("Unable to create window!");
    return (FALSE);
  };

  ShowWindow(ViewWindow, nCmdShow);
  UpdateWindow(ViewWindow);

 /*
  * Load the bitmap file specified on the command-line.  If no bitmap is
  * specified, lpCmdLine will be NULL and OpenBitmapFile() will pop up
  * a filename dialog...
  */

  OpenBitmapFile(lpCmdLine);

 /*
  * Loop on events until the user quits this application...
  */

  while (GetMessage(&msg, NULL, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  };

  return (msg.wParam);
}


/*
 * 'DisplayErrorMessage()' - Display an error message dialog.
 */

void
DisplayErrorMessage(char *format,	/* I - printf() style format string */
                    ...)		/* I - Other arguments as necessary */
{
  va_list	ap;			/* Argument pointer */
  char		s[1024];		/* Output string */


  if (format == NULL)
    return;

  va_start(ap, format);
  vsprintf(s, format, ap);
  va_end(ap);

  MessageBeep(MB_ICONEXCLAMATION);
  MessageBox(NULL, s, "Error", MB_OK | MB_ICONEXCLAMATION);
}


/*
 * 'MakePalette()' - Make a color palette for RGB colors if necessary.
 */

void
MakePalette(int pf)			/* I - Pixel format ID */
{
  PIXELFORMATDESCRIPTOR pfd;		/* Pixel format information */
  LOGPALETTE		*pPal;		/* Pointer to logical palette */
  int			nColors;	/* Number of entries in palette */
  int			i,		/* Color index */
			rmax,		/* Maximum red value */
			gmax,		/* Maximum green value */
			bmax;		/* Maximum blue value */


 /*
  * Find out if we need to define a color palette...
  */

  DescribePixelFormat(ViewDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

  if (!(pfd.dwFlags & PFD_NEED_PALETTE))
  {
    ViewPalette = NULL;
    return;
  };

 /*
  * Allocate memory for a color palette...
  */

  nColors = 1 << pfd.cColorBits;	

  pPal = (LOGPALETTE *)malloc(sizeof(LOGPALETTE) +
                              nColors * sizeof(PALETTEENTRY));

  pPal->palVersion    = 0x300;
  pPal->palNumEntries = nColors;

 /*
  * Get the maximum values for red, green, and blue.  Then build 'nColors'
  * colors...
  */

  rmax = (1 << pfd.cRedBits) - 1;
  gmax = (1 << pfd.cGreenBits) - 1;
  bmax = (1 << pfd.cBlueBits) - 1;

  for (i = 0; i < nColors; i ++)
  {
    pPal->palPalEntry[i].peRed   = 255 * ((i >> pfd.cRedShift) & rmax) / rmax;
    pPal->palPalEntry[i].peGreen = 255 * ((i >> pfd.cGreenShift) & gmax) / gmax;
    pPal->palPalEntry[i].peBlue  = 255 * ((i >> pfd.cBlueShift) & bmax) / bmax;

    pPal->palPalEntry[i].peFlags = 0;
  };
		
 /*
  * Create, select, and realize the palette...
  */

  ViewPalette = CreatePalette(pPal);
  SelectPalette(ViewDC, ViewPalette, FALSE);
  RealizePalette(ViewDC);

  free(pPal);
}


/*
 * 'ViewProc()' - Handle window events in the viewing window.
 */

LRESULT CALLBACK
ViewProc(HWND   hWnd,		/* I - Window triggering this event */
	 UINT   uMsg,		/* I - Message type */
	 WPARAM wParam,		/* I - 'word' parameter value */
	 LPARAM lParam)		/* I - 'long' parameter value */
{
  int			pf;	/* Pixel format ID */
  PIXELFORMATDESCRIPTOR pfd;	/* Pixel format information */
  PAINTSTRUCT		ps;	/* WM_PAINT message info */
  RECT			rect;	/* Current client area rectangle */


  switch (uMsg)
  {
    case WM_CREATE :
       /*
        * 'Create' message.  Get device and rendering contexts, and
	* setup the client area for OpenGL drawing...
	*/

	ViewDC = GetDC(hWnd);              

        pfd.nSize        = sizeof(pfd);
        pfd.nVersion     = 1;
        pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
						/* Do OpenGL drawing */
        pfd.dwLayerMask  = PFD_MAIN_PLANE;	/* Main drawing plane */
        pfd.iPixelType   = PFD_TYPE_RGBA;	/* RGB color buffer */
        pfd.cColorBits   = 0;			/* Best color buffer please */
        pfd.cDepthBits   = 0;			/* No depth buffer */
        pfd.cStencilBits = 0;			/* No stencil buffer */
        pfd.cAccumBits   = 0;			/* No accumulation buffer */

        pf = ChoosePixelFormat(ViewDC, &pfd);
	if (pf == 0)
          DisplayErrorMessage("oglview was unable to choose a suitable pixel format!");
        else if (!SetPixelFormat(ViewDC, pf, &pfd))
          DisplayErrorMessage("oglview was unable to set the pixel format!");

        MakePalette(pf);

        ViewRC = wglCreateContext(ViewDC);
        wglMakeCurrent(ViewDC, ViewRC);
	break;

    case WM_SIZE :
    case WM_PAINT :
       /*
        * Repaint the client area with our bitmap...
	*/

	BeginPaint(hWnd, &ps);

        GetClientRect(hWnd, &rect);
	RepaintWindow(&rect);

	EndPaint(hWnd, &ps);
	break;

    case WM_COMMAND :
       /*
        * Handle menu selections...
	*/

	switch (LOWORD(wParam))
	{
	  case IDM_FILE_OPEN :
	      OpenBitmapFile(NULL);
	      break;
	  case IDM_FILE_SAVEAS :
	      SaveBitmapFile();
	      break;
	  case IDM_FILE_PRINT :
	      PrintBitmap();
	      break;
	  case IDM_FILE_EXIT :
	      DestroyWindow(ViewWindow);
	      break;
	};
        break;

    case WM_QUIT :
    case WM_CLOSE :
       /*
        * Destroy the viewing window and exit...
        */
        
        DestroyWindow(ViewWindow);
	exit(0);
        break;

    case WM_DESTROY :
       /*
        * Release and free the device context and rendering context
	*/
  	if (ViewRC)
    	  wglDeleteContext(ViewRC);

    	if (ViewDC)
    	  ReleaseDC(ViewWindow, ViewDC);

	if (ViewPalette)
	  DeleteObject(ViewPalette);

        PostQuitMessage(0);
        break;

    case WM_QUERYNEWPALETTE :
       /*
        * Realize the color palette if necessary...
	*/

        if (ViewPalette)
	{
          SelectPalette(ViewDC, ViewPalette, FALSE);
          RealizePalette(ViewDC);

          InvalidateRect(hWnd, NULL, FALSE);
	  return (TRUE);
	};
        break;
	
    case WM_PALETTECHANGED:
       /*
        * Reselect our color palette if necessary...
	*/

	if (ViewPalette && (HWND)wParam != hWnd)
	{
          SelectPalette(ViewDC, ViewPalette, FALSE);
          RealizePalette(ViewDC);

          UpdateColors(ViewDC);
	};
	break;

    default :
       /*
        * Pass all other messages through the default window
        * procedure...
        */

        return (DefWindowProc(hWnd, uMsg, wParam, lParam));
  };

  return (FALSE);
}


/*
 * 'OpenBitmapFile()' - Load a bitmap file from disk.
 *
 * A filename dialog is popped if 'filename' is a NULL or blank string.
 */

void
OpenBitmapFile(char *filename)		/* I - Bitmap file to open */
{
  char		title[256],		/* Title of file (no path info) */
		name[256],		/* Full pathname of file */
		directory[256];		/* Directory of file */
  OPENFILENAME	ofn;			/* Filename dialog structure */
  void		*bits;			/* Bitmap pixel bits */


 /*
  * Pop up a filename dialog if we don't have a filename...
  */

  if (filename == NULL ||
      filename[0] == '\0')
  {
   /*
    * Fill in the OPENFILENAME dialog info...
    */

    strcpy(directory, ".");
    strcpy(name, "");
    strcpy(title, "");

    memset(&ofn, 0, sizeof(ofn));

    ofn.lStructSize      = sizeof(ofn);
    ofn.hwndOwner        = ViewWindow;
    ofn.lpstrFilter      = "Bitmaps\0*.BMP\0\0";
    ofn.nFilterIndex     = 1;
    ofn.lpstrFile        = name;
    ofn.nMaxFile         = sizeof(name) - 1;
    ofn.lpstrFileTitle   = title;
    ofn.nMaxFileTitle    = sizeof(title) - 1;
    ofn.lpstrInitialDir  = directory;
    ofn.lpstrTitle       = "Open Bitmap File";
    ofn.Flags            = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
                           OFN_NONETWORKBUTTON | OFN_FILEMUSTEXIST;

    if (!GetOpenFileName(&ofn))
      return;

    filename = name;
  };

 /*
  * Free the old bitmap if there is one...
  */

  if (BitmapInfo != NULL)
  {
    free(BitmapInfo);
    free(BitmapBits);
    BitmapInfo = NULL;
  };

 /*
  * Load the named bitmap from disk.  Display an error message if it
  * doesn't load...
  */

  bits = LoadDIBitmap(filename, &BitmapInfo);

  if (bits == NULL)
    DisplayErrorMessage("Could not open bitmap file \'%s\'!", filename);
  else
  {
    BitmapBits = ConvertRGB(BitmapInfo, bits);
    free(bits);

    if (BitmapBits == NULL)
    {
      DisplayErrorMessage("Could not convert bitmap to RGB!");
      free(BitmapInfo);
      BitmapInfo = NULL;
    };
  };

 /*
  * Resize the window and force a repaint...
  */

  SetWindowText(ViewWindow, filename);
  SetWindowPos(ViewWindow, HWND_TOP, 0, 0,
               BitmapInfo->bmiHeader.biWidth,
               BitmapInfo->bmiHeader.biHeight + 40,
	       SWP_NOMOVE);
}


/*
 * 'SaveBitmapFile()' - Save the currently displayed bitmap to disk.
 */

void
SaveBitmapFile(void)
{
  char		title[256],
		filename[256],
		directory[256];
  OPENFILENAME	ofn;
  void		*bits;
  BITMAPINFO	*info;


 /*
  * Grab the screen bitmap...
  */

  bits = ReadDIBitmap(&info);
  if (bits == NULL)
  {
    DisplayErrorMessage("Unable to get OpenGL bitmap from screen!");
    return;
  };

 /*
  * Pop up a filename dialog...
  */

  strcpy(directory, ".");
  strcpy(filename, "untitled.bmp");
  strcpy(title, "");

  memset(&ofn, 0, sizeof(ofn));

  ofn.lStructSize      = sizeof(ofn);
  ofn.hwndOwner        = ViewWindow;
  ofn.lpstrFilter      = "Bitmaps\0*.BMP\0\0";
  ofn.nFilterIndex     = 1;
  ofn.lpstrFile        = filename;
  ofn.nMaxFile         = sizeof(filename) - 1;
  ofn.lpstrFileTitle   = title;
  ofn.nMaxFileTitle    = sizeof(title) - 1;
  ofn.lpstrInitialDir  = directory;
  ofn.lpstrTitle       = "Save Bitmap File";
  ofn.Flags            = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
                         OFN_NONETWORKBUTTON;

  if (GetSaveFileName(&ofn))
  {
   /*
    * Save the named bitmap to disk...
    */

    if (SaveDIBitmap(filename, info, bits))
      DisplayErrorMessage("Could not save to file \'%s\' -\n%s",
                          filename, strerror(errno));
  };

 /*
  * Free memory and return...
  */

  free(info);
  free(bits);
}


/*
 * 'PrintBitmap()' - Print the currently displayed bitmap.
 */

void
PrintBitmap(void)
{
  void		*bits;
  BITMAPINFO	*info;


 /*
  * Grab the screen bitmap...
  */

  bits = ReadDIBitmap(&info);
  if (bits == NULL)
  {
    DisplayErrorMessage("Unable to get OpenGL bitmap from screen!");
    return;
  };

 /*
  * Try printing it...
  */

  PrintDIBitmap(ViewWindow, info, bits);

 /*
  * Free memory and return...
  */

  free(info);
  free(bits);
}


/*
 * 'RepaintWindow()' - Redraw the client area with our bitmap...
 */

void
RepaintWindow(RECT        *rect)	/* I - Client area rectangle */
{
  GLint		xoffset,		/* X offset of image */
  		yoffset;		/* Y offset of image */
  GLint		xsize,			/* X size of scaled image */
		ysize;			/* Y size of scaled image */
  GLfloat	xscale,			/* Scaling in X direction */
		yscale;			/* Scaling in Y direction */


 /*
  * Reset the viewport and clear the window to white...
  */

  glViewport(0, 0, rect->right, rect->bottom);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, rect->right - 1.0, 0.0, rect->bottom - 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

 /*
  * If we have loaded a bitmap image, scale it to fit the window...
  */

  if (BitmapBits != NULL)
  {
    xsize = rect->right;
    ysize = BitmapInfo->bmiHeader.biHeight * xsize /
            BitmapInfo->bmiHeader.biWidth;
    if (ysize > rect->bottom)
    {
      ysize = rect->bottom;
      xsize = BitmapInfo->bmiHeader.biWidth * ysize /
              BitmapInfo->bmiHeader.biHeight;
    };

    xscale  = (float)xsize / (float)BitmapInfo->bmiHeader.biWidth;
    yscale  = (float)ysize / (float)BitmapInfo->bmiHeader.biHeight;

    xoffset = (rect->right - xsize) * 0.5;
    yoffset = (rect->bottom - ysize) * 0.5;

    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    glPixelZoom(xscale, yscale);
    glRasterPos2i(xoffset, yoffset);
    glDrawPixels(BitmapInfo->bmiHeader.biWidth,
                 BitmapInfo->bmiHeader.biHeight,
		 GL_RGB, GL_UNSIGNED_BYTE, BitmapBits);
  };

  glFinish();
}


/*
 * End of "$Id: oglview.c,v 1.1 1995/12/31 07:27:56 mike Exp mike $".
 */
