// map.c
// Draws a relief map of North America

#include <windows.h>            // Window defines
#include <gl\gl.h>              // OpenGL
#include <gl\glu.h>             // GLU library
#include <gl\glaux.h>
#include "resource.h"           // About box resource identifiers.


const char *g_szFileName = "data\\usa.bin";


#define MAP_COLUMNS 1477
#define MAP_ROWS 841
#define MAP_LIST 1000
#define GRID_SIZE 100	// How many meters per 5 minutes of lat/long

HPALETTE hPalette = NULL;


// Application name and instance storeage
static LPCTSTR lpszAppName = "Relief Map";
static HINSTANCE hInstance;

// Declaration for Window procedure
LRESULT CALLBACK WndProc(   HWND    hWnd,
							UINT    message,
							WPARAM  wParam,
							LPARAM  lParam);

// Dialog procedure for about box
BOOL APIENTRY AboutDlgProc (HWND hDlg, UINT message, UINT wParam, LONG lParam);

// Set Pixel Format function - forward declaration
void SetDCPixelFormat(HDC hDC);



// Reduces a normal vector specified as a set of three coordinates,
// to a unit normal vector of length one.
void ReduceToUnitd(float vector[3])
	{
	float length;
	
	// Calculate the length of the vector		
	length = (float)sqrt((vector[0]*vector[0]) + 
						(vector[1]*vector[1]) +
						(vector[2]*vector[2]));

	// Keep the program from blowing up by providing an exceptable
	// value for vectors that may calculated too close to zero.
	if(length == 0.0f)
		length = 1.0f;

	// Dividing each element by the length will result in a
	// unit normal vector.
	vector[0] /= length;
	vector[1] /= length;
	vector[2] /= length;
	}


// Points p1, p2, & p3 specified in counter clock-wise order
void calcNormald(float v[3][3], float out[3])
	{
	float v1[3],v2[3];
	static const int x = 0;
	static const int y = 1;
	static const int z = 2;

	// Calculate two vectors from the three points
	v1[x] = v[0][x] - v[1][x];
	v1[y] = v[0][y] - v[1][y];
	v1[z] = v[0][z] - v[1][z];

	v2[x] = v[1][x] - v[2][x];
	v2[y] = v[1][y] - v[2][y];
	v2[z] = v[1][z] - v[2][z];

	// Take the cross product of the two vectors to get
	// the normal vector which will be stored in out
	out[x] = v1[y]*v2[z] - v1[z]*v2[y];
	out[y] = v1[z]*v2[x] - v1[x]*v2[z];
	out[z] = v1[x]*v2[y] - v1[y]*v2[x];

	// Normalize the vector (shorten length to one)
	ReduceToUnitd(out);
	}



// Read in two rows and use the data to build a triangle
// mesh.
BOOL bReadMapData(short *MapData)
	{
	HANDLE hFile;
	DWORD cb;
	int i;

	// Open the file for reading, optimize for sequential access
	hFile = CreateFile(g_szFileName,GENERIC_READ,0,NULL,
		OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,NULL);

	// Check for I/O errors
	if(hFile == INVALID_HANDLE_VALUE)
		{
		MessageBox(NULL,"Cannot open map file!","File Error",MB_OK | MB_ICONSTOP);
		return FALSE;
		}

	
	// Read in in one big flush
	if( (!ReadFile(hFile,MapData,sizeof(short)*MAP_COLUMNS*MAP_ROWS,&cb,NULL)) ||
		(cb != MAP_COLUMNS*MAP_ROWS*sizeof(short)) )
		{
		MessageBox(NULL,"Error reading data!","File Error",MB_OK | MB_ICONSTOP);
		CloseHandle(hFile);
		return FALSE;
		}

	// Close the data file
	CloseHandle(hFile);

	// Shrink the elevation data
	for(i = 0; i < MAP_COLUMNS*MAP_ROWS; i++)
		MapData[i] = (short)((float)(MapData[i])*0.075f);

	return TRUE;
	}




// Change viewing volume and viewport.  Called when window is resized
void ChangeSize(GLsizei w, GLsizei h)
	{
	GLfloat nRange = (GLfloat)(MAP_COLUMNS/2+25);

	// Prevent a divide by zero
	if(h == 0)
		h = 1;

	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);

	// Reset coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Establish clipping volume (left, right, bottom, top, near, far)
    if (w <= h) 
		glOrtho (-nRange, nRange, nRange*h/w, -nRange*h/w, -nRange*4.0f, nRange*4.0f);
    else 
		glOrtho (-nRange*w/h, nRange*w/h, nRange, -nRange, -nRange*4.0f, nRange*4.0f);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	}



// This function does any needed initialization on the rendering
// context.  Here it sets up and initializes the lighting for
// the scene.
void SetupRC()
	{
	// Light values and coordinates
	GLfloat  whiteLight[] = {0.7f, 0.7f, 0.7f, 1.0f };
	GLfloat  diffuse[] = { 0.9f, 0.9f, 0.9f, 1.0f};
	GLfloat	 lightPos[] = { 3000.0f, 0.0f, -6000.0f, 1.0f };


	glEnable(GL_DEPTH_TEST);	// Hidden surface removal
	glFrontFace(GL_CCW);		// Counter clockwise polygons are front
	glCullFace(GL_BACK);		// Front will be counter clockwise face
	glEnable(GL_CULL_FACE);		// Do not calculate inside of solid object
	
	
	// Enable lighting
	glEnable(GL_LIGHTING);

	// Setup light 0
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,whiteLight);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);

	// Position and turn on the light
	glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
	glEnable(GL_LIGHT0);

	// Enable color tracking
	glEnable(GL_COLOR_MATERIAL);
	
	// Set Material properties to follow glColor values
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

	// Black background
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f );
	}



// Take the map data and build the display list mesh
void BuildMesh(short *MapData)
	{
	int step = 5; // 5 gives good results
	int lat,lng;
	int latOffset,lngOffset;
	float vertex[3][3];
	float normal[3];

	glNewList(MAP_LIST,GL_COMPILE);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glBegin(GL_TRIANGLES);


	latOffset = MAP_ROWS/2;
	lngOffset = MAP_COLUMNS/2;

	for(lat = 0; lat < MAP_ROWS-step; lat+=(1*step))
		for(lng = 0; lng < MAP_COLUMNS-step; lng+=(1*step))
			{
			// Top triangle
			// First Point
			vertex[0][0] = (float)(lng - lngOffset);
			vertex[0][1] = (float)(lat - lngOffset);
			vertex[0][2] = (float)MapData[((lat*MAP_COLUMNS)+lng)];
			
			// Second Point
			vertex[1][0] = (float)(lng+step-lngOffset);
			vertex[1][1] = (float)(lat+step-lngOffset);
			vertex[1][2] = (float)(MapData[(((lat+step)*MAP_COLUMNS)+lng+step)]);

			// Third Point
			vertex[2][0] = (float)(lng+(2*step)-lngOffset);
			vertex[2][1] = (float)(lat-lngOffset);
			vertex[2][2] = (float)(MapData[((lat*MAP_COLUMNS)+lng+step+step)]);

			// Calculate Normal
			calcNormald(vertex,normal);

			// Specidy the normal/triangle
			glNormal3fv(normal);

			// Blue id below sea level
			if(vertex[0][2] > 0.0)
				glColor3d(0.60, 0.4, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[0]);


			if(vertex[1][2] > 0.0)
				glColor3d(0.60, 0.4, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[1]);


			if(vertex[2][2] > 0.0)
				glColor3d(0.60, 0.4, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[2]);


			// Bottom triangle
			// First Point
			vertex[0][0] = (float)(lng+(2*step)-lngOffset);
			vertex[0][1] = (float)(lat-lngOffset);
			vertex[0][2] = (float)(MapData[((lat*MAP_COLUMNS)+lng+step+step)]);

			// Second Point
			vertex[1][0] = (float)(lng+step-lngOffset);
			vertex[1][1] = (float)(lat+step-lngOffset);
			vertex[1][2] = (float)(MapData[(((lat+step)*MAP_COLUMNS)+lng+step)]);
			
			// Third Point
			vertex[2][0] = (float)(lng+(3*step)-lngOffset);
			vertex[2][1] = (float)(lat+(1*step)-lngOffset);
			vertex[2][2] = (float)(MapData[(((lat+step)*MAP_COLUMNS)+lng+(3*step))]);
			
			// Calculate Normal
			calcNormald(vertex,normal);

			// Specidy the normal/triangle
			glNormal3fv(normal);

			// Blue id below sea level
			if(vertex[0][2] > 0.0)
				glColor3d(0.60, 0.40, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[0]);

			if(vertex[1][2] > 0.0)
				glColor3d(0.60, 0.4, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[1]);

			if(vertex[2][2] > 0.0)
				glColor3d(0.60, 0.4, 0.12);
			else
				glColor3d(0.0, 0.0, 1.0);

			glVertex3fv(vertex[2]);
			}

	glEnd();

	glEndList();
	}



// Called to draw scene
void RenderScene(void)
	{
	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Save the matrix state (MODELVIEW)
	glMatrixMode(GL_MODELVIEW);
	
	glPushMatrix();
	
	glTranslatef(20.0f, (1500.0f-(float)MAP_ROWS)/2.0f, 0.0f);

	// Render relief map
	glCallList(MAP_LIST);

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Flush drawing commands
	glFlush();
	}


// Select the pixel format for a given device context
void SetDCPixelFormat(HDC hDC)
	{
	int nPixelFormat;

	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),  // Size of this structure
		1,                                                              // Version of this structure    
		PFD_DRAW_TO_WINDOW |                    // Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |					// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,                       // Double buffered
		PFD_TYPE_RGBA,                          // RGBA Color mode
		24,                                     // Want 24bit color 
		0,0,0,0,0,0,                            // Not used to select mode
		0,0,                                    // Not used to select mode
		0,0,0,0,0,                              // Not used to select mode
		16,                                     // Size of depth buffer
		0,                                      // Not used to select mode
		0,                                      // Not used to select mode
		PFD_MAIN_PLANE,                         // Draw in main plane
		0,                                      // Not used to select mode
		0,0,0 };                                // Not used to select mode

	// Choose a pixel format that best matches that described in pfd
	nPixelFormat = ChoosePixelFormat(hDC, &pfd);

	// Set the pixel format for the device context
	SetPixelFormat(hDC, nPixelFormat, &pfd);
	}



// If necessary, creates a 3-3-2 palette for the device context listed.
HPALETTE GetOpenGLPalette(HDC hDC)
	{
	HPALETTE hRetPal = NULL;	// Handle to palette to be created
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;						// Counting variable
	BYTE RedRange,GreenRange,BlueRange;
								// Range for each color entry (7,7,and 3)


	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return NULL
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return NULL;

	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;	

	// Allocate space for a logical palette structure plus all the palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +nColors*sizeof(PALETTEENTRY));

	// Fill in palette header 
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; // table size

	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.  
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;

	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
		{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);

		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) & GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);

		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);

		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
		}
		

	// Create the palette
	hRetPal = CreatePalette(pPal);

	// Go ahead and select and realize the palette for this device context
	SelectPalette(hDC,hRetPal,FALSE);
	RealizePalette(hDC);

	// Free the memory used for the logical palette structure
	free(pPal);

	// Return the handle to the new palette
	return hRetPal;
	}


// Entry point of all Windows programs
int APIENTRY WinMain(   HINSTANCE       hInst,
						HINSTANCE       hPrevInstance,
						LPSTR           lpCmdLine,
						int                     nCmdShow)
	{
	MSG                     msg;            // Windows message structure
	WNDCLASS        wc;                     // Windows class structure
	HWND            hWnd;           // Storeage for window handle

	hInstance = hInst;

	// Register Window style
	wc.style                        = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc          = (WNDPROC) WndProc;
	wc.cbClsExtra           = 0;
	wc.cbWndExtra           = 0;
	wc.hInstance            = hInstance;
	wc.hIcon                        = NULL;
	wc.hCursor                      = LoadCursor(NULL, IDC_ARROW);
	
	// No need for background brush for OpenGL window
	wc.hbrBackground        = NULL;         
	
	wc.lpszMenuName         = MAKEINTRESOURCE(IDR_MENU);
	wc.lpszClassName        = lpszAppName;

	// Register the window class
	if(RegisterClass(&wc) == 0)
		return FALSE;


	// Create the main application window
	hWnd = CreateWindow(
				lpszAppName,
				lpszAppName,
				
				// OpenGL requires WS_CLIPCHILDREN and WS_CLIPSIBLINGS
				WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
	
				// Window position and size
				50, 50,
				400, 400,
				NULL,
				NULL,
				hInstance,
				NULL);

	// If window was not created, quit
	if(hWnd == NULL)
		return FALSE;


	// Display the window
	ShowWindow(hWnd,SW_SHOW);
	UpdateWindow(hWnd);

	// Process application messages until the application closes
	while( GetMessage(&msg, NULL, 0, 0))
		{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		}

	return msg.wParam;
	}




// Window procedure, handles all messages for this program
LRESULT CALLBACK WndProc(   HWND    hWnd,
							UINT    message,
							WPARAM  wParam,
							LPARAM  lParam)
	{
	static HGLRC hRC;               // Permenant Rendering context
	static HDC hDC;                 // Private GDI Device context
	static short *MapData;		    // Destination for map data
	int memReq;

	switch (message)
		{
		// Window creation, setup for OpenGL
		case WM_CREATE:
			// Allocate enough memory to contain map data
			memReq = sizeof(short)*MAP_COLUMNS*MAP_ROWS;
			MapData = malloc(memReq);

			// Check for error allocating memory for map data
			if(MapData == NULL)
				{
				MessageBox(hWnd,"Cannot allocate memory for Map Data!","Memory Error",MB_OK | MB_ICONSTOP);
				PostQuitMessage(0);
				return 0L;
				}

			// Read the data in
			if(!bReadMapData(MapData))
				{
				MessageBox(hWnd,"Could not read map data!","File Error",MB_OK | MB_ICONSTOP);
				free(MapData);
				PostQuitMessage(0);
				return 0L;
				}

			// Store the device context
			hDC = GetDC(hWnd);              

			// Select the pixel format
			SetDCPixelFormat(hDC);          

			hPalette = GetOpenGLPalette(hDC);

			// Create the rendering context and make it current
			hRC = wglCreateContext(hDC);
			wglMakeCurrent(hDC, hRC);
			SetupRC();
		
			// Builds the display list for the triangle mesh
			BuildMesh(MapData);


			// Free the map data
			free(MapData);

			break;

		// Window is being destroyed, cleanup
		case WM_DESTROY:
			// Deselect the current rendering context and delete it
			wglMakeCurrent(hDC,NULL);
			wglDeleteContext(hRC);

			if(hPalette != NULL)
				DeleteObject(hPalette);

			// Tell the application to terminate after the window
			// is gone.
			PostQuitMessage(0);
			break;

		// Window is resized.
		case WM_SIZE:
			// Call our function which modifies the clipping
			// volume and viewport
			ChangeSize(LOWORD(lParam), HIWORD(lParam));
			break;


		// The painting function.  This message sent by Windows 
		// whenever the screen needs updating.
		case WM_PAINT:
			{
			// Call OpenGL drawing code
			RenderScene();

			SwapBuffers(hDC);

			// Validate the newly painted client area
			ValidateRect(hWnd,NULL);
			}
			break;

		// Windows is telling the application that it may modify
		// the system palette.  This message in essance asks the 
		// application for a new palette.
		case WM_QUERYNEWPALETTE:
			// If the palette was created.
			if(hPalette)
				{
				int nRet;

				// Selects the palette into the current device context
				SelectPalette(hDC, hPalette, FALSE);

				// Map entries from the currently selected palette to
				// the system palette.  The return value is the number 
				// of palette entries modified.
				nRet = RealizePalette(hDC);

				// Repaint, forces remap of palette in current window
				InvalidateRect(hWnd,NULL,FALSE);

				return nRet;
				}
			break;

	
		// This window may set the palette, even though it is not the 
		// currently active window.
		case WM_PALETTECHANGED:
			// Don't do anything if the palette does not exist, or if
			// this is the window that changed the palette.
			if((hPalette != NULL) && ((HWND)wParam != hWnd))
				{
				// Select the palette into the device context
				SelectPalette(hDC,hPalette,FALSE);

				// Map entries to system palette
				RealizePalette(hDC);
				
				// Remap the current colors to the newly realized palette
				UpdateColors(hDC);
				return 0;
				}
			break;


		// A menu command
		case WM_COMMAND:
			{
			switch(LOWORD(wParam))
				{
				// Exit the program
				case ID_FILE_EXIT:
					DestroyWindow(hWnd);
					break;

				// Display the about box
				case ID_HELP_ABOUT:
					DialogBox (hInstance,
						MAKEINTRESOURCE(IDD_DIALOG_ABOUT),
						hWnd,
						AboutDlgProc);
					break;
				}
			}
			break;


	default:   // Passes it on if unproccessed
	    return (DefWindowProc(hWnd, message, wParam, lParam));

	}

    return (0L);
	}




// Dialog procedure.
BOOL APIENTRY AboutDlgProc (HWND hDlg, UINT message, UINT wParam, LONG lParam)
	{
	
    switch (message)
	{
		// Initialize the dialog box
	    case WM_INITDIALOG:
			{
			int i;
			GLenum glError;

			// glGetString demo
			SetDlgItemText(hDlg,IDC_OPENGL_VENDOR,glGetString(GL_VENDOR));
			SetDlgItemText(hDlg,IDC_OPENGL_RENDERER,glGetString(GL_RENDERER));
			SetDlgItemText(hDlg,IDC_OPENGL_VERSION,glGetString(GL_VERSION));
			SetDlgItemText(hDlg,IDC_OPENGL_EXTENSIONS,glGetString(GL_EXTENSIONS));

			// gluGetString demo
			SetDlgItemText(hDlg,IDC_GLU_VERSION,gluGetString(GLU_VERSION));
			SetDlgItemText(hDlg,IDC_GLU_EXTENSIONS,gluGetString(GLU_EXTENSIONS));


			// Display any recent error messages
			i = 0;
			do {
				glError = glGetError();
				SetDlgItemText(hDlg,IDC_ERROR1+i,gluErrorString(glError));
				i++;
				}
			while(i < 6 && glError != GL_NO_ERROR);


			return (TRUE);
			}
			break;

		// Process command messages
	    case WM_COMMAND:      
			{
			// Validate and Make the changes
			if(LOWORD(wParam) == IDOK)
				EndDialog(hDlg,TRUE);
		    }
			break;

		// Closed from sysbox
		case WM_CLOSE:
			EndDialog(hDlg,TRUE);
			break;
		}

	return FALSE;
	}
