// glcode.c
// Demonstrates use of Vertex and Normal arrays

//Visteon (TM) and its logo are trademarks of Visteon Corporation and may not
//be used without the express written permission of the company. Such
//permission has been granted to Richard Wright for use limited to this 
//publication. Copying, altering, or further re-creation or use for any
//reason by other parties is strictly prohibited.


// Requisite headers
#include <windows.h>
#include <math.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glaux.h>
#include "glcode.h"


// Vertex Array definitions and storage
typedef struct _VERTEX {
	GLfloat	fVx;
	GLfloat fVy;
	GLfloat fVz;
	} VERTEX;



typedef struct _NORMAL {
	GLfloat fNx;
	GLfloat fNy;
	GLfloat fNz;
	} NORMAL;


NORMAL NORMALS[368];
VERTEX VERTICES[368];

// Do a Dual load of the vertex and normal arrays
void LoadVArray(int nIndex, GLfloat v[3], GLfloat n[3])
	{
	VERTICES[nIndex].fVx = v[0];
	VERTICES[nIndex].fVy = v[1];
	VERTICES[nIndex].fVz = v[2];

	NORMALS[nIndex].fNx = n[0];
	NORMALS[nIndex].fNy = n[1];
	NORMALS[nIndex].fNz = n[2];
	}

int g_nSidesIndex = 264;
int g_nLogoFront;
int g_nEnd;

// Pieces Parts
void DrawSphere(GLfloat fRadius);
void DrawFloor(void);
void DrawVisteon(void);
void DrawAllButFloor(BOOL fShadow);
void MakeShadowMatrix(GLfloat points[3][3], GLfloat lightPos[4], GLfloat destMat[4][4]);


void BuildFloor();
void BuildSides();
void BuildFaces(GLfloat face[16][3], GLfloat back[16][3]);

// Easier way to specify color for Windows bigots
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)

////////////////////////////////////////////////////
// Globals
////////////////////////////////////////////////////

// Lighting components /////////////////////////////
GLfloat  ambientLight[] = { 0.5f, 0.5f, 0.5f, 1.0f };
GLfloat  diffuseLight[] = { 0.6f, 0.6f, 0.6f, 1.0f };
GLfloat  specular[] = { 1.0f, 1.0f, 1.0f, 1.0f};
GLfloat	 lightPos[] = { -75.0f, 200.0f, 100.0f, 1.0f };
GLfloat  spotDir[] = { 0.3f, -1.0f, -0.3f };

GLfloat shadowMat[4][4];
GLfloat floorPoints[3][3] ={{ 0.0f, -74.0f, 0.0f },
							{ 1.0f, -74.0f, 0.0f },
							{ 1.0f, -74.0f, -1.0f }};


////////////////////////////////////////////////////
GLfloat fYRot = 0.0f;

// Display list for Background
GLuint		StaticList;
GLuint		textList;
GLuint		miscList;
GLuint		logoList;

/////////////////////////////////////////////////////////////
// GENERAL UTILITY ROUTINES
/////////////////////////////////////////////////////////////
// Reduces a normal vector specified as a set of three coordinates,
// to a unit normal vector of length one.
void ReduceToUnit(float vector[3])
	{
	float length;
	
	// Calculate the length of the vector		
	length = (float)sqrt((vector[0]*vector[0]) + 
						(vector[1]*vector[1]) +
						(vector[2]*vector[2]));

	// Keep the program from blowing up by providing an exceptable
	// value for vectors that may calculated too close to zero.
	if(length == 0.0f)
		length = 1.0f;

	// Dividing each element by the length will result in a
	// unit normal vector.
	vector[0] /= length;
	vector[1] /= length;
	vector[2] /= length;
	}


// Points p1, p2, & p3 specified in counter clock-wise order
void calcNormal(float v[3][3], float out[3])
	{
	float v1[3],v2[3];
	static const int x = 0;
	static const int y = 1;
	static const int z = 2;

	// Calculate two vectors from the three points
	v1[x] = v[0][x] - v[1][x];
	v1[y] = v[0][y] - v[1][y];
	v1[z] = v[0][z] - v[1][z];

	v2[x] = v[1][x] - v[2][x];
	v2[y] = v[1][y] - v[2][y];
	v2[z] = v[1][z] - v[2][z];

	// Take the cross product of the two vectors to get
	// the normal vector which will be stored in out
	out[x] = v1[y]*v2[z] - v1[z]*v2[y];
	out[y] = v1[z]*v2[x] - v1[x]*v2[z];
	out[z] = v1[x]*v2[y] - v1[y]*v2[x];

	// Normalize the vector (shorten length to one)
	ReduceToUnit(out);
	}



// Creates a shadow projection matrix out of the plane equation
// coefficients and the position of the light. The return value is stored
// in destMat[][]
void MakeShadowMatrix(GLfloat points[3][3], GLfloat lightPos[4], GLfloat destMat[4][4])
	{
	GLfloat planeCoeff[4];
	GLfloat dot;

	// Find the plane equation coefficients
	// Find the first three coefficients the same way we
	// find a normal.
	calcNormal(points,planeCoeff);

	// Find the last coefficient by back substitutions
	planeCoeff[3] = - (
		(planeCoeff[0]*points[2][0]) + (planeCoeff[1]*points[2][1]) +
		(planeCoeff[2]*points[2][2]));


	// Dot product of plane and light position
	dot = planeCoeff[0] * lightPos[0] +
			planeCoeff[1] * lightPos[1] +
			planeCoeff[2] * lightPos[2] +
			planeCoeff[3] * lightPos[3];

	// Now do the projection
	// First column
    destMat[0][0] = dot - lightPos[0] * planeCoeff[0];
    destMat[1][0] = 0.0f - lightPos[0] * planeCoeff[1];
    destMat[2][0] = 0.0f - lightPos[0] * planeCoeff[2];
    destMat[3][0] = 0.0f - lightPos[0] * planeCoeff[3];

	// Second column
	destMat[0][1] = 0.0f - lightPos[1] * planeCoeff[0];
	destMat[1][1] = dot - lightPos[1] * planeCoeff[1];
	destMat[2][1] = 0.0f - lightPos[1] * planeCoeff[2];
	destMat[3][1] = 0.0f - lightPos[1] * planeCoeff[3];

	// Third Column
	destMat[0][2] = 0.0f - lightPos[2] * planeCoeff[0];
	destMat[1][2] = 0.0f - lightPos[2] * planeCoeff[1];
	destMat[2][2] = dot - lightPos[2] * planeCoeff[2];
	destMat[3][2] = 0.0f - lightPos[2] * planeCoeff[3];

	// Fourth Column
	destMat[0][3] = 0.0f - lightPos[3] * planeCoeff[0];
	destMat[1][3] = 0.0f - lightPos[3] * planeCoeff[1];
	destMat[2][3] = 0.0f - lightPos[3] * planeCoeff[2];
	destMat[3][3] = dot - lightPos[3] * planeCoeff[3];
	}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Do any initialization of the rendering context here, such as
// setting background colors, setting up lighting, or performing
// preliminary calculations.
void GLSetupRC(void *pData)
	{
	HDC hDC;
	HFONT hFont;
	GLYPHMETRICSFLOAT agmf[256]; // Throw away
	LOGFONT logfont;

	logfont.lfHeight = -10;
	logfont.lfWidth = 0;
	logfont.lfEscapement = 0;
	logfont.lfOrientation = 0;
	logfont.lfWeight = FW_BOLD;
	logfont.lfItalic = FALSE;
	logfont.lfUnderline = FALSE;
	logfont.lfStrikeOut = FALSE;
	logfont.lfCharSet = ANSI_CHARSET;
	logfont.lfOutPrecision = OUT_DEFAULT_PRECIS;
	logfont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	logfont.lfQuality = DEFAULT_QUALITY;
	logfont.lfPitchAndFamily = DEFAULT_PITCH;
	strcpy(logfont.lfFaceName,"Arial");


	glEnable(GL_DEPTH_TEST);	// Hidden surface removal
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_COLOR_MATERIAL);

	// Kill the back faces
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);

	glClearColor(0.2f, 0.2f, 0.2f, 1.0f );

	glEnable(GL_LIGHTING);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuseLight);
	glLightfv(GL_LIGHT0,GL_SPECULAR,specular);
	glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
	glLightf(GL_LIGHT0,GL_SPOT_CUTOFF,90.0f);
	
	// Fairly shiny spot
	glLightf(GL_LIGHT0,GL_SPOT_EXPONENT,20.0f);	
	glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,spotDir);
	glEnable(GL_LIGHT0);


	glEnable(GL_NORMALIZE);
	glEnable(GL_AUTO_NORMAL);

	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glMaterialfv(GL_FRONT, GL_SPECULAR,specular);
	glMateriali(GL_FRONT,GL_SHININESS,10);

	hDC = (HDC)pData;
	hFont = CreateFontIndirect(&logfont);
	SelectObject (hDC, hFont); 

	
	//create display lists for glyphs 0 through 255 with 0.1 extrusion 
	// and default deviation. The display list numbering starts at 1000 
	// (it could be any number). 
	wglUseFontOutlines(hDC, 0, 255, 10000, 0.0f, 0.3f, 
				WGL_FONT_POLYGONS, agmf); 

	DeleteObject(hFont);

	// Prebuild the shadow matrix
	MakeShadowMatrix(floorPoints, lightPos, shadowMat);

	/////////////////////////////////////////////////////////////////////////////
	// OpenGL version 1.1 ///////////////////////////////////////////////////////
	// Get Ready for the Vertex Array's
    glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	BuildFloor();

	// Setup Arrays
	glVertexPointer(3, GL_FLOAT, sizeof(VERTEX), &VERTICES[0].fVx);
	glNormalPointer(GL_FLOAT, sizeof(NORMAL), &NORMALS[0].fNx);


	// Build all the display lists
	StaticList = glGenLists(10);
	textList = StaticList + 1;
	miscList = textList + 1;
	logoList = miscList + 1;
	
	glNewList(StaticList,GL_COMPILE);
		// Clear the window with current clearing color
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// Draw the floor and back plane, which do not move

		// Save the vertex array status, disable the use of the Normal Arrays
		// then draw the floor and re-enable use of Normal Arrays.
		// Why: On large flat surfaces, it is faster (your code executes quicker)
		// to specify the normal once and let it affect each subsequent vertex than
		// to specify it specifically for each vertex.
		glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
		glDisableClientState(GL_NORMAL_ARRAY);
		DrawFloor();
		glPopClientAttrib();
	glEndList();


	glNewList(textList,GL_COMPILE);
		// Put Text and bar
		// Save state before placing text
		glPushMatrix();

		// Place the text /////////////////////
		glListBase(10000);
		glTranslatef(-30.0f,-50.0f,2.5f);
		glScalef(15.0f, 15.0f, 15.0f);

		glCallLists(7, GL_UNSIGNED_BYTE, "VISTEON");
		glPopMatrix();

		glPushMatrix();
		glTranslatef(-30.0f, -65.0f, 1.5f);
		glScalef(9.0f, 9.0f, 9.0f);

		glCallLists(11, GL_UNSIGNED_BYTE, "CORPORATION");
		glPopMatrix();
	glEndList();

	glNewList(miscList,GL_COMPILE);
		// Draw the sphere 
		glPushMatrix();
		glTranslatef(0.0f, 70.0f, 0.0f);
		DrawSphere(7.5f);
		glPopMatrix();

		// Draw the border between VISTEON and CORPORATION
		glPushMatrix();
		glTranslatef(0.0f, -54.0f, 0.0f);
		auxSolidBox(60.0f, 2.0f, 2.0f);
		glPopMatrix();
	glEndList();


	glNewList(logoList,GL_COMPILE);
		glPopMatrix();
		DrawVisteon();
		glPopMatrix();
	glEndList();
 	}



//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Render your OpenGL Scene here.
void GLRenderScene(void *pData)
	{
	// Update Positional stuff
	fYRot += 5.0f;

	if(fYRot > 360.0f)
		fYRot = 0.0f;

	// Draw the floor and background image
    glCallList(StaticList);

	// Draw the spinning stuff
	DrawAllButFloor(FALSE);

	glPushMatrix();
	glMultMatrixf((const GLfloat*)shadowMat);

	DrawAllButFloor(TRUE);

	glPopMatrix();

	// Flush drawing commands
	glFlush();
	}


void DrawAllButFloor(BOOL fShadow)
	{
	// Ebony
	if(fShadow)
		glRGB(0,0,0);
	else
		glRGB(192, 192, 192);

	glPushMatrix();

	// Rotate the viewing transformation
	// Rotate around Y Axis
	glRotatef(fYRot,0.0f, 1.0f, 0.0f); 

	glCallList(textList);

	// Bright RED
	if(fShadow)
		glRGB(0,0,0);
	else
		glColor3f(0.9f, 0.5f, 0.0f);

	glCallList(miscList);
		
	// Rotated, but still at origin
	if(fShadow)
		glRGB(0,0,0);
	else
		glColor3f(0.2f, 0.2f, 0.2f);

	glCallList(logoList);
	}

////////////////////////////////////////////////////////////
// Pieces Parts

// Fills the vertex and normal array elements with the data needed to build
// the floor.
void BuildFloor()
	{
	int nIndex = 0;

	GLfloat fFloorSize = 175.0f;
	GLfloat fDivision = 35.0f;
	GLfloat fX, fZ;
	GLfloat fStep = fFloorSize/fDivision;

	// Fill in the vertices for the QUAD Strips
	for(fX = -fFloorSize-fDivision; fX < fFloorSize+fDivision; fX += fDivision)
		{
		VERTICES[nIndex].fVx = fX;
		VERTICES[nIndex].fVy = -75.0f;
		VERTICES[nIndex].fVz = -fFloorSize;
		nIndex++;
		VERTICES[nIndex].fVx = fX+fDivision;
		VERTICES[nIndex].fVy = -75.0f;
		VERTICES[nIndex].fVz = -fFloorSize;
		nIndex++;
		for(fZ = -fFloorSize+fDivision; fZ < fFloorSize+fDivision; fZ += fDivision)
			{
			VERTICES[nIndex].fVx = fX;
			VERTICES[nIndex].fVy = -75.0f;
			VERTICES[nIndex].fVz = -fZ;
			nIndex++;
			VERTICES[nIndex].fVx = fX+fDivision;
			VERTICES[nIndex].fVy = -75.0f;
			VERTICES[nIndex].fVz = -fZ;
			nIndex++;
			}
		}
	}


// The floor must be positioned first
void DrawFloor(void)
	{
	int i;

	// White Tile
	glRGB(255,255,255);
	glNormal3f(0.0f, 1.0f, 0.0f);
	
	for(i = 0; i < 12; i++)
		glDrawArrays(GL_QUAD_STRIP, i*22, 22);
	}


// Draw the sphere, the sphere must be positioned first!
void DrawSphere(GLfloat fRadius)
	{	
	GLUquadricObj *pQObj;

	pQObj = gluNewQuadric();

	gluQuadricDrawStyle(pQObj, GLU_FILL);
	gluQuadricNormals(pQObj, GLU_SMOOTH);

	// Switch front face
	glFrontFace(GL_CCW);

	// object, radius, slices, stacks
	gluSphere(pQObj, fRadius, 20, 20);

	gluDeleteQuadric(pQObj);
	}





// For speed, we will build the entire logo rather than do all
// the translations, etc. that we could do
// It would be faster still if we ditch the scaleing....
void DrawVisteon()
	{
	GLfloat fScale = 0.5f;
	GLfloat back[16][3];
	GLfloat fWidth = 7.0f;
	int i;

	// Vertices of front face. All else is derived from these
	GLfloat front[16][3] = { { 15.0f, 172.0f, 0.0f }, // 0
							  { 82.0f, 172.0f, 0.0f }, // 1
							  { 72.0f, 155.0f,  0.0f }, // 2
							  { 23.0f, 155.0f,  0.0f }, // 3
							  { 23.0f, 148.0f,  0.0f }, // 4
							  { 67.0f, 148.0f,  0.0f }, // 5
							  { 57.2f, 131.0f,  0.0f }, // 6
							  { 23.0f, 131.0f,  0.0f }, // 7
							  { 23.0f, 124.0f,  0.0f }, // 8
							  { 52.0f, 124.0f,  0.0f }, // 9
							  { 42.0f, 107.0f,  0.0f }, // 10 
							  { 23.0f, 107.0f,  0.0f }, // 11
							  { 23.0f, 100.0f,  0.0f }, // 12
							  { 38.0f,  100.0f,  0.0f }, // 13
							  { 23.0f,  76.0f,  0.0f }, // 14
							  { 0.0f,  0.0f,  0.0f }}; // 15
		
	// Build the front and back
	for(i = 0; i < 16; i++)
		{
		back[i][0] = front[i][0];
		back[i][1] = front[i][1];
		back[i][2] = -fWidth;
		front[i][2] = fWidth;
		}
	
	BuildSides(front, back);
	BuildFaces(front, back);

	glTranslatef(0.0f, -30.0f, 0.0f);
	glScalef(fScale, fScale, fScale);

	// Draw the Faces and sides
	glDrawArrays(GL_QUADS, g_nLogoFront, (g_nEnd - g_nLogoFront)+1);
	glDrawArrays(GL_QUADS, g_nSidesIndex, (g_nLogoFront-g_nSidesIndex)+1);
	
	// Draw the left side
	glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
	
	// Draw the faces and sides
	glDrawArrays(GL_QUADS, g_nLogoFront, (g_nEnd - g_nLogoFront)+1);
	glDrawArrays(GL_QUADS, g_nSidesIndex, (g_nLogoFront-g_nSidesIndex)+1);
	}



//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Called by Windows when it receives the WM_SIZE message.
// Put any code needed here to recalc the viewing volume and
// viewport info
void GLResize(GLsizei w, GLsizei h)
	{
	GLfloat nRange = 100.0f;
	GLfloat fAspect;

	// Prevent a divide by zero
	if(h == 0)
		h = 1;

	fAspect = (GLfloat)w/(GLfloat)h;

	// Set Viewport to window dimensions
	glViewport(0, 0, w, h);

	glMatrixMode(GL_PROJECTION);

	// Reset coordinate system
	glLoadIdentity();

	// Setup perspective for viewing
	gluPerspective(45.0f,fAspect,100,700);

	// Viewing transformation
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -300.0f);

	// Reposition the lights now that we have reset the transformation
	glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
	glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,spotDir);
	}




// Close up the sides
void BuildSides(GLfloat front[16][3], GLfloat back[16][3])
	{
	int nIndex = g_nSidesIndex;
	GLfloat fNormal[3];
	GLfloat fPoints[3][3];

	// All the Top Sections
	fNormal[0] = 0.0f;
	fNormal[1] = 1.0f;
	fNormal[2] = 0.0f;
	
	LoadVArray(nIndex++, front[0], fNormal);
	LoadVArray(nIndex++, front[1], fNormal);
	LoadVArray(nIndex++, back[1], fNormal);
	LoadVArray(nIndex++, back[0], fNormal);

	LoadVArray(nIndex++, front[4], fNormal);
	LoadVArray(nIndex++, front[5], fNormal);
	LoadVArray(nIndex++, back[5], fNormal);
	LoadVArray(nIndex++, back[4], fNormal);


	LoadVArray(nIndex++, front[8], fNormal);
	LoadVArray(nIndex++, front[9], fNormal);
	LoadVArray(nIndex++, back[9], fNormal);
	LoadVArray(nIndex++, back[8], fNormal);


	LoadVArray(nIndex++, front[12], fNormal);
	LoadVArray(nIndex++, front[13], fNormal);
	LoadVArray(nIndex++, back[13], fNormal);
	LoadVArray(nIndex++, back[12], fNormal);


	// All the bottom Secctions
	fNormal[0] = 0.0f;
	fNormal[1] = -1.0f;
	fNormal[2] = 0.0f;

	LoadVArray(nIndex++, front[2], fNormal);
	LoadVArray(nIndex++, front[3], fNormal);
	LoadVArray(nIndex++, back[3], fNormal);
	LoadVArray(nIndex++, back[2], fNormal);


	LoadVArray(nIndex++, front[6], fNormal);
	LoadVArray(nIndex++, front[7], fNormal);
	LoadVArray(nIndex++, back[7], fNormal);
	LoadVArray(nIndex++, back[6], fNormal);


	LoadVArray(nIndex++, front[10], fNormal);
	LoadVArray(nIndex++, front[11], fNormal);
	LoadVArray(nIndex++, back[11], fNormal);
	LoadVArray(nIndex++, back[10], fNormal);


	// Side panels with normals pointing to the right
	fNormal[0] = 1.0f;
	fNormal[1] = 0.0f;
	fNormal[2] = 0.0f;


	LoadVArray(nIndex++, front[3], fNormal);
	LoadVArray(nIndex++, front[4], fNormal);
	LoadVArray(nIndex++, back[4], fNormal);
	LoadVArray(nIndex++, back[3], fNormal);

	LoadVArray(nIndex++, front[7], fNormal);
	LoadVArray(nIndex++, front[8], fNormal);
	LoadVArray(nIndex++, back[8], fNormal);
	LoadVArray(nIndex++, back[7], fNormal);

	LoadVArray(nIndex++, front[11], fNormal);
	LoadVArray(nIndex++, front[12], fNormal);
	LoadVArray(nIndex++, back[12], fNormal);
	LoadVArray(nIndex++, back[11], fNormal);

	// Now do the slanted faces that need normals
	// calculate
	fPoints[0][0] = front[1][0];
	fPoints[0][1] = front[1][1];
	fPoints[0][2] = front[1][2];

	fPoints[1][0] = front[2][0];
	fPoints[1][1] = front[2][1];
	fPoints[1][2] = front[2][2];

	fPoints[2][0] = back[2][0];
	fPoints[2][1] = back[2][1];
	fPoints[2][2] = back[2][2];
	
	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[1], fNormal);
	LoadVArray(nIndex++, front[2], fNormal);
	LoadVArray(nIndex++, back[2], fNormal);
	LoadVArray(nIndex++, back[1], fNormal);


	fPoints[0][0] = front[5][0];
	fPoints[0][1] = front[5][1];
	fPoints[0][2] = front[5][2];

	fPoints[1][0] = front[6][0];
	fPoints[1][1] = front[6][1];
	fPoints[1][2] = front[6][2];

	fPoints[2][0] = back[6][0];
	fPoints[2][1] = back[6][1];
	fPoints[2][2] = back[6][2];

	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[5], fNormal);
	LoadVArray(nIndex++, front[6], fNormal);
	LoadVArray(nIndex++, back[6], fNormal);
	LoadVArray(nIndex++, back[5], fNormal);


	fPoints[0][0] = front[13][0];
	fPoints[0][1] = front[13][1];
	fPoints[0][2] = front[13][2];

	fPoints[1][0] = front[14][0];
	fPoints[1][1] = front[14][1];
	fPoints[1][2] = front[14][2];

	fPoints[2][0] = back[14][0];
	fPoints[2][1] = back[14][1];
	fPoints[2][2] = back[14][2];

	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[9], fNormal);
	LoadVArray(nIndex++, front[10], fNormal);
	LoadVArray(nIndex++, back[10], fNormal);
	LoadVArray(nIndex++, back[9], fNormal);


	fPoints[0][0] = front[13][0];
	fPoints[0][1] = front[13][1];
	fPoints[0][2] = front[13][2];

	fPoints[1][0] = front[14][0];
	fPoints[1][1] = front[14][1];
	fPoints[1][2] = front[14][2];

	fPoints[2][0] = back[14][0];
	fPoints[2][1] = back[14][1];
	fPoints[2][2] = back[14][2];

	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[13], fNormal);
	LoadVArray(nIndex++, front[14], fNormal);
	LoadVArray(nIndex++, back[14], fNormal);
	LoadVArray(nIndex++, back[13], fNormal);


	fPoints[0][0] = front[0][0];
	fPoints[0][1] = front[0][1];
	fPoints[0][2] = front[0][2];

	fPoints[1][0] = back[0][0];
	fPoints[1][1] = back[0][1];
	fPoints[1][2] = back[0][2];

	fPoints[2][0] = back[15][0];
	fPoints[2][1] = back[15][1];
	fPoints[2][2] = back[15][2];

	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[0], fNormal);
	LoadVArray(nIndex++, back[0], fNormal);
	LoadVArray(nIndex++, back[15], fNormal);
	LoadVArray(nIndex++, front[15], fNormal);


	fPoints[0][0] = front[14][0];
	fPoints[0][1] = front[14][1];
	fPoints[0][2] = front[14][2];

	fPoints[1][0] = front[15][0];
	fPoints[1][1] = front[15][1];
	fPoints[1][2] = front[15][2];

	fPoints[2][0] = back[15][0];
	fPoints[2][1] = back[15][1];
	fPoints[2][2] = back[15][2];

	calcNormal(fPoints,fNormal);

	LoadVArray(nIndex++, front[14], fNormal);
	LoadVArray(nIndex++, front[15], fNormal);
	LoadVArray(nIndex++, back[15], fNormal);
	LoadVArray(nIndex++, back[14], fNormal);

	g_nLogoFront = nIndex;
	}



void BuildFaces(GLfloat face[16][3], GLfloat back[16][3])
	{
	GLfloat fNormal[3];
	int nIndex = g_nLogoFront;

	fNormal[0] = 0.0f;
	fNormal[1] = 0.0f;
	fNormal[2] = 1.0f;

	// Build Front Facing Face first
	LoadVArray(nIndex++, face[0], fNormal);
	LoadVArray(nIndex++, face[15], fNormal);
	LoadVArray(nIndex++, face[14], fNormal);
	LoadVArray(nIndex++, face[3], fNormal);

	LoadVArray(nIndex++, face[0], fNormal);
	LoadVArray(nIndex++, face[3], fNormal);
	LoadVArray(nIndex++, face[2], fNormal);
	LoadVArray(nIndex++, face[1], fNormal);

	LoadVArray(nIndex++, face[5], fNormal);
	LoadVArray(nIndex++, face[4], fNormal);
	LoadVArray(nIndex++, face[7], fNormal);
	LoadVArray(nIndex++, face[6], fNormal);

	LoadVArray(nIndex++, face[9], fNormal);
	LoadVArray(nIndex++, face[8], fNormal);
	LoadVArray(nIndex++, face[11], fNormal);
	LoadVArray(nIndex++, face[10], fNormal);

	LoadVArray(nIndex++, face[13], fNormal);
	LoadVArray(nIndex++, face[12], fNormal);
	LoadVArray(nIndex++, face[14], fNormal);
	LoadVArray(nIndex++, face[14], fNormal);


	// Now build the back faces, same as above, 
	fNormal[2] = -1.0f;
	LoadVArray(nIndex++, back[3], fNormal);
	LoadVArray(nIndex++, back[14], fNormal);
	LoadVArray(nIndex++, back[15], fNormal);
	LoadVArray(nIndex++, back[0], fNormal);

	LoadVArray(nIndex++, back[1], fNormal);
	LoadVArray(nIndex++, back[2], fNormal);
	LoadVArray(nIndex++, back[3], fNormal);
	LoadVArray(nIndex++, back[0], fNormal);

	LoadVArray(nIndex++, back[6], fNormal);
	LoadVArray(nIndex++, back[7], fNormal);
	LoadVArray(nIndex++, back[4], fNormal);
	LoadVArray(nIndex++, back[5], fNormal);

	LoadVArray(nIndex++, back[10], fNormal);
	LoadVArray(nIndex++, back[11], fNormal);
	LoadVArray(nIndex++, back[8], fNormal);
	LoadVArray(nIndex++, back[9], fNormal);

	LoadVArray(nIndex++, back[14], fNormal);
	LoadVArray(nIndex++, back[14], fNormal);
	LoadVArray(nIndex++, back[12], fNormal);
	LoadVArray(nIndex++, back[13], fNormal);
	g_nEnd = nIndex;
	}
