//----------------------------------------------------------------------------
//  Project Oglview
//  Waite Group Press
//  Copyright  1996. All Rights Reserved.
//
//  SUBSYSTEM:    Oglview Application
//  FILE:         oglvwwnd.cpp
//  AUTHOR:       Richard S. Wright Jr.
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TOglviewWindow (TWindow).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>
#include "oglvwapp.h"
#include "oglvwwnd.h"
#include "bitmap.h"
#include <stdio.h>


// Forward declarations
void GLResize(GLsizei w, GLsizei h);
void GLRenderScene(void *pData);
void GLSetupRC(void *pData);

// Easier way to specify color for Windows bigots
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)


HWND		ViewWindow;		/* View window */
HPALETTE	ViewPalette;		/* Color palette (if necessary) */
HDC		ViewDC;			/* Drawing context */
HGLRC		ViewRC;			/* OpenGL rendering context */


BITMAPINFO	*BitmapInfo = NULL;	/* Current bitmap information */
void		*BitmapBits = NULL;	/* Current bitmap pixel bits */

// May not need all of these
void			DisplayErrorMessage(char *, ...);
void			MakePalette(int);
LRESULT CALLBACK	ViewProc(HWND, UINT, WPARAM, LPARAM);
void			OpenBitmapFile(char *);
void			SaveBitmapFile(void);
void			PrintBitmap(void);
void			RepaintWindow(RECT *);


//
// Build a response table for all messages/commands handled by the application.
//
DEFINE_RESPONSE_TABLE1(TOglviewWindow, TWindow)
//{{TOglviewWindowRSP_TBL_BEGIN}}
  EV_COMMAND(CM_FILEOPEN, CmFileopen),
  EV_COMMAND(CM_FILESAVEAS, CmFilesaveas),
  EV_WM_CREATE,
  EV_WM_DESTROY,
  EV_WM_ERASEBKGND,
  EV_WM_PAINT,
  EV_WM_SIZE,
  EV_WM_QUERYNEWPALETTE,
  EV_WM_PALETTECHANGED,
  EV_COMMAND(CM_PRINT, CmFileitem1),
//{{TOglviewWindowRSP_TBL_END}}
END_RESPONSE_TABLE;


//{{TOglviewWindow Implementation}}


//--------------------------------------------------------
// TOglviewWindow
// ~~~~~~~~~~
// Construction/Destruction handling.
//
TOglviewWindow::TOglviewWindow(TWindow* parent, const char far* title, TModule* module)
:
  TWindow(parent, title, module)
{
  	m_pPalette = NULL;
	m_hDC = NULL;
}


TOglviewWindow::~TOglviewWindow()
{
	Destroy();
  	delete m_pPalette;
}

void TOglviewWindow::CmFileopen()
{
	OpenBitmapFile(NULL);
}


void TOglviewWindow::CmFilesaveas()
{
	SaveBitmapFile();
}


void TOglviewWindow::CmFileitem1()
{
    PrintBitmap();
}


int TOglviewWindow::EvCreate(CREATESTRUCT far& createStruct)
{
  int result;

  createStruct.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);

  result = TWindow::EvCreate(createStruct);

  	// Select pixel format/rendering context
	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,					// Version of this structure
		PFD_DRAW_TO_WINDOW |	// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |	// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,		// Double buffered mode
		PFD_TYPE_RGBA,		// RGBA Color mode
		24,				// Want 24bit color
		0,0,0,0,0,0,			// Not used to select mode
		0,0,				// Not used to select mode
		0,0,0,0,0,			// Not used to select mode
		32,				// Size of depth buffer
		0,				// Not used to select mode
		0,				// Not used to select mode
		PFD_MAIN_PLANE,		// Draw in main plane
		0,				// Not used to select mode
		0,0,0 };			// Not used to select mode

	// Get the device context
	m_hDC = ::GetDC(this->GetHandle());

	// Choose a pixel format that best matches that described in pfd
	int nPixelFormat = ChoosePixelFormat(m_hDC, &pfd);

	// Set the pixel format for the device context
	SetPixelFormat(m_hDC, nPixelFormat, &pfd);

	// Create a 3-3-2 palette
   SetupPalette(m_hDC);

	// Create the rendering context
	m_hRC = wglCreateContext(m_hDC);

	// Make the rendering context current and perform initializion.
	wglMakeCurrent(m_hDC,m_hRC);

	ViewWindow = this->HWindow;		/* View window */
	ViewPalette = (HPALETTE)(*m_pPalette);		/* Color palette (if necessary) */
	ViewDC = m_hDC;			/* Drawing context */
	ViewRC = m_hRC;			/* OpenGL rendering context */

   return result;
}


void TOglviewWindow::EvDestroy()
{
   // Free the rendering context
	wglMakeCurrent(NULL,NULL);
   wglDeleteContext(m_hRC);

	// Release the device context
   ::ReleaseDC(this->GetHandle(),m_hDC);

   TWindow::EvDestroy();
}


bool TOglviewWindow::EvEraseBkgnd(HDC dc)
{
	return FALSE;
}


void TOglviewWindow::EvPaint()
{
	// Make the rendering context current, and call OpenGL Rendering code
	wglMakeCurrent(m_hDC,m_hRC);
	GLRenderScene(NULL);
	wglMakeCurrent(NULL,m_hRC);

	// Finally swap buffers since this rendering context is double buffered
	SwapBuffers(m_hDC);

	// Validate the window
	Validate();
}


void TOglviewWindow::EvSize(uint sizeType, TSize& size)
{
    TWindow::EvSize(sizeType, size);
}


bool TOglviewWindow::EvQueryNewPalette()
{
	bool result;

	// Only if palette was created
	if(m_pPalette != NULL)
		{
		int nRet;

		// Select the palette into the current device context
		if(SelectPalette(m_hDC, m_pPalette->GetHandle(),FALSE) == NULL)
			::MessageBox(NULL,"Cannot select Palette","Error",MB_OK);

		// Map entries from the currently selected palette to
		// the system palette.  The return value is the number
		// of palette entries modified.
		nRet = RealizePalette(m_hDC);

		// Repaint, forces remap of palette in current window
		Invalidate();

		return nRet;
		}

	// Call default function
	result = TWindow::EvQueryNewPalette();
	return result;
}


void TOglviewWindow::EvPaletteChanged(THandle hWndPalChg)
{
	// Only if palette created, or not this window
	if((m_pPalette != NULL) && (hWndPalChg != this->HWindow))
		{
		// Select the palette into the device context
		::SelectPalette(m_hDC,m_pPalette->GetHandle(),FALSE);

		// Map entries to system palette
		::RealizePalette(m_hDC);

		// Remap the current colors to the newly realized palette
		::UpdateColors(m_hDC);
		return;
		}

	// Call default handler
	TWindow::EvPaletteChanged(hWndPalChg);
}

// Create the palette if necessary
void TOglviewWindow::SetupPalette(HDC hDC)
	{
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;					// Counting variable

	BYTE RedRange,GreenRange,BlueRange;  // Range for each color entrys
						    // 			(7,7,and 3)

	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat,
		sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return;

	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;

	// Allocate space for a logical palette structure plus all the
	// palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE)
	+nColors*sizeof(PALETTEENTRY));

	// Fill in palette header
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; 	// table size

	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;

	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
		{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);

		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) &
										GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);

		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);

		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
		}

	// Create the palette
	m_pPalette = new TPalette(pPal);

	// Go ahead and select and realize the palette for this device context
	if(SelectPalette(hDC,m_pPalette->GetHandle(),FALSE) == NULL)
		::MessageBox(NULL,"Cannot select Palette in Palette Creation",
								   "Error",MB_OK);

	if(RealizePalette(hDC) == NULL)
		::MessageBox(NULL,"Cannot realize Palette in Palette Creation",
									"Error",MB_OK);

	// Free the memory used for the logical palette structure
	free(pPal);
	}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Render your OpenGL Scene here.
void GLRenderScene(void *pData)
	{
	RECT rect;

    GetClientRect(ViewWindow, &rect);
	RepaintWindow(&rect);
	}




/*
 * 'OpenBitmapFile()' - Load a bitmap file from disk.
 *
 * A filename dialog is popped if 'filename' is a NULL or blank string.
 */

void
OpenBitmapFile(char *filename)		/* I - Bitmap file to open */
{
  char		title[256],		/* Title of file (no path info) */
		name[256],		/* Full pathname of file */
		directory[256];		/* Directory of file */
  OPENFILENAME	ofn;			/* Filename dialog structure */
  void		*bits;			/* Bitmap pixel bits */


 /*
  * Pop up a filename dialog if we don't have a filename...
  */

  if (filename == NULL ||
      filename[0] == '\0')
  {
   /*
    * Fill in the OPENFILENAME dialog info...
    */

    strcpy(directory, ".");
    strcpy(name, "");
    strcpy(title, "");

    memset(&ofn, 0, sizeof(ofn));

    ofn.lStructSize      = sizeof(ofn);
    ofn.hwndOwner        = ViewWindow;
    ofn.lpstrFilter      = "Bitmaps\0*.BMP\0\0";
    ofn.nFilterIndex     = 1;
    ofn.lpstrFile        = name;
    ofn.nMaxFile         = sizeof(name) - 1;
    ofn.lpstrFileTitle   = title;
    ofn.nMaxFileTitle    = sizeof(title) - 1;
    ofn.lpstrInitialDir  = directory;
    ofn.lpstrTitle       = "Open Bitmap File";
    ofn.Flags            = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
                           OFN_NONETWORKBUTTON | OFN_FILEMUSTEXIST;

    if (!GetOpenFileName(&ofn))
      return;

    filename = name;
  };

 /*
  * Free the old bitmap if there is one...
  */

  if (BitmapInfo != NULL)
  {
    free(BitmapInfo);
    free(BitmapBits);
    BitmapInfo = NULL;
  };

 /*
  * Load the named bitmap from disk.  Display an error message if it
  * doesn't load...
  */

  bits = LoadDIBitmap(filename, &BitmapInfo);

  if (bits == NULL)
    {
    DisplayErrorMessage("Could not open bitmap file \'%s\'!", filename);
    return;
    }
  else
  {
    BitmapBits = ConvertRGB(BitmapInfo, bits);
    free(bits);

    if (BitmapBits == NULL)
    {
      DisplayErrorMessage("Could not convert bitmap to RGB!");
      free(BitmapInfo);
      BitmapInfo = NULL;
    };
  };

 /*
  * Resize the window and force a repaint...
  */

  SetWindowText(ViewWindow, filename);
  SetWindowPos(ViewWindow, HWND_TOP, 0, 0,
               BitmapInfo->bmiHeader.biWidth,
               BitmapInfo->bmiHeader.biHeight + 40,
	       SWP_NOMOVE);
}


/*
 * 'SaveBitmapFile()' - Save the currently displayed bitmap to disk.
 */

void
SaveBitmapFile(void)
{
  char		title[256],
		filename[256],
		directory[256];
  OPENFILENAME	ofn;
  void		*bits;
  BITMAPINFO	*info;


 /*
  * Grab the screen bitmap...
  */

  bits = ReadDIBitmap(&info);
  if (bits == NULL)
  {
    DisplayErrorMessage("Unable to get OpenGL bitmap from screen!");
    return;
  };

 /*
  * Pop up a filename dialog...
  */

  strcpy(directory, ".");
  strcpy(filename, "untitled.bmp");
  strcpy(title, "");

  memset(&ofn, 0, sizeof(ofn));

  ofn.lStructSize      = sizeof(ofn);
  ofn.hwndOwner        = ViewWindow;
  ofn.lpstrFilter      = "Bitmaps\0*.BMP\0\0";
  ofn.nFilterIndex     = 1;
  ofn.lpstrFile        = filename;
  ofn.nMaxFile         = sizeof(filename) - 1;
  ofn.lpstrFileTitle   = title;
  ofn.nMaxFileTitle    = sizeof(title) - 1;
  ofn.lpstrInitialDir  = directory;
  ofn.lpstrTitle       = "Save Bitmap File";
  ofn.Flags            = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST |
                         OFN_NONETWORKBUTTON;

  if (GetSaveFileName(&ofn))
  {
   /*
    * Save the named bitmap to disk...
    */

    if (SaveDIBitmap(filename, info, bits))
      DisplayErrorMessage("Could not save to file \'%s\' -\n%s",
                          filename, strerror(errno));
  };

 /*
  * Free memory and return...
  */

  free(info);
  free(bits);
}


/*
 * 'PrintBitmap()' - Print the currently displayed bitmap.
 */

void
PrintBitmap(void)
{
  void		*bits;
  BITMAPINFO	*info;


 /*
  * Grab the screen bitmap...
  */

  bits = ReadDIBitmap(&info);
  if (bits == NULL)
  {
    DisplayErrorMessage("Unable to get OpenGL bitmap from screen!");
    return;
  };

 /*
  * Try printing it...
  */

  PrintDIBitmap(ViewWindow, info, bits);

 /*
  * Free memory and return...
  */

  free(info);
  free(bits);
}


/*
 * 'RepaintWindow()' - Redraw the client area with our bitmap...
 */

void
RepaintWindow(RECT        *rect)	/* I - Client area rectangle */
{
  GLint		xoffset,		/* X offset of image */
  		yoffset;		/* Y offset of image */
  GLint		xsize,			/* X size of scaled image */
		ysize;			/* Y size of scaled image */
  GLfloat	xscale,			/* Scaling in X direction */
		yscale;			/* Scaling in Y direction */


 /*
  * Reset the viewport and clear the window to white...
  */

  glViewport(0, 0, rect->right, rect->bottom);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, rect->right - 1.0, 0.0, rect->bottom - 1.0, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);

  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

 /*
  * If we have loaded a bitmap image, scale it to fit the window...
  */

  if (BitmapBits != NULL)
  {
    xsize = rect->right;
    ysize = BitmapInfo->bmiHeader.biHeight * xsize /
            BitmapInfo->bmiHeader.biWidth;
    if (ysize > rect->bottom)
    {
      ysize = rect->bottom;
      xsize = BitmapInfo->bmiHeader.biWidth * ysize /
              BitmapInfo->bmiHeader.biHeight;
    };

    xscale  = (float)xsize / (float)BitmapInfo->bmiHeader.biWidth;
    yscale  = (float)ysize / (float)BitmapInfo->bmiHeader.biHeight;

    xoffset = (rect->right - xsize) * 0.5;
    yoffset = (rect->bottom - ysize) * 0.5;

    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    glPixelZoom(xscale, yscale);
    glRasterPos2i(xoffset, yoffset);
    glDrawPixels(BitmapInfo->bmiHeader.biWidth,
                 BitmapInfo->bmiHeader.biHeight,
		 GL_RGB, GL_UNSIGNED_BYTE, BitmapBits);
  };

  glFinish();
}


/*
 * 'DisplayErrorMessage()' - Display an error message dialog.
 */

void
DisplayErrorMessage(char *format,	/* I - printf() style format string */
                    ...)		/* I - Other arguments as necessary */
{
  va_list	ap;			/* Argument pointer */
  char		s[1024];		/* Output string */


  if (format == NULL)
    return;

  va_start(ap, format);
  vsprintf(s, format, ap);
  va_end(ap);

  MessageBeep(MB_ICONEXCLAMATION);
  MessageBox(NULL, s, "Error", MB_OK | MB_ICONEXCLAMATION);
}


