//----------------------------------------------------------------------------
//  Project Nurbs
//  Waite Group Press
//  Copyright  1996. All Rights Reserved.
//
//  SUBSYSTEM:    Nurbs Application
//  FILE:         nrbswndw.cpp
//  AUTHOR:       Richard S. Wright Jr.
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TNurbsWindow (TWindow).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>
#include "nurbsapp.h"
#include "nrbswndw.h"
#include <stdio.h>


// Forward declarations
void GLResize(GLsizei w, GLsizei h);
void GLRenderScene(void *pData);
void GLSetupRC(void *pData);

// Easier way to specify color for Windows bigots
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)


// NURBS object pointer
GLUnurbsObj *pNurb = NULL;

// The number of control points for this curve
GLint nNumPoints = 4; // 4 X 4

// Mesh extends four units -6 to +6 along x and y axis
// Lies in Z plane
//                 u  v  (x,y,z)	
GLfloat ctrlPoints[4][4][3]= {{{  -6.0f, -6.0f, 0.0f},	// u = 0,	v = 0
						      {	  -6.0f, -2.0f, 0.0f},	//			v = 1
							  {   -6.0f,  2.0f, 0.0f},	//			v = 2	
							  {   -6.0f,  6.0f, 0.0f}}, //			v = 3
							 
							  {{  -2.0f, -6.0f, 0.0f},	// u = 1	v = 0
						      {   -2.0f, -2.0f, 8.0f},	//			v = 1
							  {   -2.0f,  2.0f, 8.0f},	//			v = 2
//						      {   -2.0f, -2.0f, 0.0f},	//			v = 1
//							  {   -2.0f,  2.0f, 0.0f},	//			v = 2
							  {   -2.0f,  6.0f, 0.0f}},	//			v = 3
							  
							  {{   2.0f, -6.0f, 0.0f }, // u =2		v = 0
							  {    2.0f, -2.0f, 8.0f }, //			v = 1
							  {    2.0f,  2.0f, 8.0f },	//			v = 2
//							  {    2.0f, -2.0f, 0.0f }, //			v = 1
//							  {    2.0f,  2.0f, 0.0f },	//			v = 2
							  {    2.0f,  6.0f, 0.0f }},//			v = 3

							  {{   6.0f, -6.0f, 0.0f},	// u = 3	v = 0
							  {    6.0f, -2.0f, 0.0f},	//			v = 1
							  {    6.0f,  2.0f, 0.0f},	//			v = 2
							  {    6.0f,  6.0f, 0.0f}}};//			v = 3

// Knot sequence for the NURB
GLfloat Knots[8] = {0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f};



// Called to draw the control points in Red over the NURB
void DrawPoints(void)
	{
	int i,j;

	// Large Red Points
	glPointSize(5.0f);
	glRGB(255,0,0);

	// Draw all the points in the array
	glBegin(GL_POINTS);
		for(i = 0; i < 4; i++)
			for(j = 0; j < 4; j++)
				glVertex3fv(ctrlPoints[i][j]);	
	glEnd();
	}



//
// Build a response table for all messages/commands handled by the application.
//
DEFINE_RESPONSE_TABLE1(TNurbsWindow, TWindow)
//{{TNurbsWindowRSP_TBL_BEGIN}}
  EV_WM_CREATE,
  EV_WM_DESTROY,
  EV_WM_ERASEBKGND,
  EV_WM_PAINT,
  EV_WM_SIZE,
  EV_WM_PALETTECHANGED,
  EV_WM_QUERYNEWPALETTE,
//{{TNurbsWindowRSP_TBL_END}}
END_RESPONSE_TABLE;


//{{TNurbsWindow Implementation}}


//--------------------------------------------------------
// TNurbsWindow
// ~~~~~~~~~~
// Construction/Destruction handling.
//
TNurbsWindow::TNurbsWindow(TWindow* parent, const char far* title, TModule* module)
:
  TWindow(parent, title, module)
{
  	m_pPalette = NULL;
	m_hDC = NULL;
}


TNurbsWindow::~TNurbsWindow()
{
	Destroy();
  	delete m_pPalette;
}

int TNurbsWindow::EvCreate(CREATESTRUCT far& createStruct)
{
  int result;

  createStruct.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);

  result = TWindow::EvCreate(createStruct);

  	// Select pixel format/rendering context
	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,					// Version of this structure
		PFD_DRAW_TO_WINDOW |	// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |	// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,		// Double buffered mode
		PFD_TYPE_RGBA,		// RGBA Color mode
		24,				// Want 24bit color
		0,0,0,0,0,0,			// Not used to select mode
		0,0,				// Not used to select mode
		0,0,0,0,0,			// Not used to select mode
		32,				// Size of depth buffer
		0,				// Not used to select mode
		0,				// Not used to select mode
		PFD_MAIN_PLANE,		// Draw in main plane
		0,				// Not used to select mode
		0,0,0 };			// Not used to select mode

	// Get the device context
	m_hDC = ::GetDC(this->GetHandle());

	// Choose a pixel format that best matches that described in pfd
	int nPixelFormat = ChoosePixelFormat(m_hDC, &pfd);

	// Set the pixel format for the device context
	SetPixelFormat(m_hDC, nPixelFormat, &pfd);

	// Create a 3-3-2 palette
   SetupPalette(m_hDC);

	// Create the rendering context
	m_hRC = wglCreateContext(m_hDC);

	// Make the rendering context current and perform initializion.
	wglMakeCurrent(m_hDC,m_hRC);
	GLSetupRC(m_hDC);

   // Set a timer for 200 milliseconds
   SetTimer(200,101,NULL);

   return result;
}


void TNurbsWindow::EvDestroy()
{
   // Free the rendering context
	wglMakeCurrent(NULL,NULL);
   wglDeleteContext(m_hRC);

	// Release the device context
   ::ReleaseDC(this->GetHandle(),m_hDC);

   TWindow::EvDestroy();
}


bool TNurbsWindow::EvEraseBkgnd(HDC dc)
{
	return FALSE;
}


void TNurbsWindow::EvPaint()
{
	// Make the rendering context current, and call OpenGL Rendering code
	wglMakeCurrent(m_hDC,m_hRC);
	GLRenderScene(NULL);
	wglMakeCurrent(NULL,m_hRC);

	// Finally swap buffers since this rendering context is double buffered
	SwapBuffers(m_hDC);

	// Validate the window
	Validate();
}


void TNurbsWindow::EvSize(uint sizeType, TSize& size)
{
    TWindow::EvSize(sizeType, size);

	// Make the rendering context current, and call function
	// to make adjustments to OpenGL viewport
	wglMakeCurrent(m_hDC,m_hRC);
	GLResize(size.cx, size.cy);
	wglMakeCurrent(m_hDC,NULL);
}


void TNurbsWindow::EvPaletteChanged(THandle hWndPalChg)
{
	// Only if palette created, or not this window
	if((m_pPalette != NULL) && (hWndPalChg != this->HWindow))
		{
		// Select the palette into the device context
		::SelectPalette(m_hDC,m_pPalette->GetHandle(),FALSE);

		// Map entries to system palette
		::RealizePalette(m_hDC);

		// Remap the current colors to the newly realized palette
		::UpdateColors(m_hDC);
		return;
		}

	// Call default handler
	TWindow::EvPaletteChanged(hWndPalChg);
}


bool TNurbsWindow::EvQueryNewPalette()
{
	bool result;

	// Only if palette was created
	if(m_pPalette != NULL)
		{
		int nRet;

		// Select the palette into the current device context
		if(SelectPalette(m_hDC, m_pPalette->GetHandle(),FALSE) == NULL)
			::MessageBox(NULL,"Cannot select Palette","Error",MB_OK);

		// Map entries from the currently selected palette to
		// the system palette.  The return value is the number
		// of palette entries modified.
		nRet = RealizePalette(m_hDC);

		// Repaint, forces remap of palette in current window
		Invalidate();

		return nRet;
		}

	// Call default function
	result = TWindow::EvQueryNewPalette();
	return result;
}


// Create the palette if necessary
void TNurbsWindow::SetupPalette(HDC hDC)
	{
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;					// Counting variable

	BYTE RedRange,GreenRange,BlueRange;  // Range for each color entrys
						    // 			(7,7,and 3)

	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat,
		sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return;

	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;

	// Allocate space for a logical palette structure plus all the
	// palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE)
	+nColors*sizeof(PALETTEENTRY));

	// Fill in palette header
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; 	// table size

	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;

	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
		{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);

		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) &
										GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);

		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);

		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
		}

	// Create the palette
	m_pPalette = new TPalette(pPal);

	// Go ahead and select and realize the palette for this device context
	if(SelectPalette(hDC,m_pPalette->GetHandle(),FALSE) == NULL)
		::MessageBox(NULL,"Cannot select Palette in Palette Creation",
								   "Error",MB_OK);

	if(RealizePalette(hDC) == NULL)
		::MessageBox(NULL,"Cannot realize Palette in Palette Creation",
									"Error",MB_OK);

	// Free the memory used for the logical palette structure
	free(pPal);
	}




//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Called by Windows when it receives the WM_SIZE message.
// Put any code needed here to recalc the viewing volume and
// viewport info
void GLResize(GLsizei w, GLsizei h)
	{
	// Prevent a divide by zero
	if(h == 0)
		h = 1;

   if(w == 0)
   	w = 1;

	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Perspective view
	gluPerspective (45.0f, (GLdouble)w/(GLdouble)h, 1.0, 40.0f);

	// Modelview matrix reset
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Viewing transformation, position for better view
    glTranslatef (0.0f, 0.0f, -20.0f);
	}

// This function does any needed initialization on the rendering
// context.  Here it sets up and initializes the lighting for
// the scene.
GLfloat	 lightPos[] = { 0.0f, 0.0f, -30.0f, 1.0f };
void GLSetupRC(void *pData)
	{
	// Light values and coordinates
	GLfloat  whiteLight[] = {0.7f, 0.7f, 0.7f, 1.0f };
	GLfloat  specular[] = { 0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat  shine[] = { 100.0f };

	// Clear Window to white
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f );

	// Enable lighting
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// Enable color tracking
	glEnable(GL_COLOR_MATERIAL);
	
	// Set Material properties to follow glColor values
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, shine);
	
	// Automatically generate normals for evaluated surfaces
	glEnable(GL_AUTO_NORMAL);


	// Setup the Nurbs object
    pNurb = gluNewNurbsRenderer();

	
    gluNurbsProperty(pNurb, GLU_SAMPLING_TOLERANCE, 25.0f);
	// Uncomment the next line and comment the one following to produce a
	// wire frame mesh.
    //gluNurbsProperty(pNurb, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
	gluNurbsProperty(pNurb, GLU_DISPLAY_MODE, (GLfloat)GLU_FILL);
	}




// Called to draw scene
void GLRenderScene(void *pData)
	{
	// Draw in Blue
	glRGB(0,0,220);

	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Save the modelview matrix stack
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// Rotate the mesh around to make it easier to see
	glRotatef(330.0f, 1.0f,0.0f,0.0f);
	
	// Render the NURB
	// Begin the NURB definition
    gluBeginSurface(pNurb);
	
	// Evaluate the surface
    gluNurbsSurface(pNurb,	// pointer to NURBS renderer
        8, Knots,			// No. of knots and knot array u direction	
        8, Knots,			// No. of knots and knot array v direction
        4 * 3,				// Distance between control points in u dir.
        3,					// Distance between control points in v dir.
        &ctrlPoints[0][0][0], // Control points
        4, 4,					// u and v order of surface
        GL_MAP2_VERTEX_3);		// Type of surface
		
	// Done with surface
	gluEndSurface(pNurb);
	
	// Show the control points
	DrawPoints();

	// Restore the modelview matrix
	glPopMatrix();

	// Flush drawing commands
	glFlush();
	}


