//----------------------------------------------------------------------------
//  Project Atom
//  Waite Group Press
//  Copyright  1996. All Rights Reserved.
//
//  SUBSYSTEM:    Atom Application
//  FILE:         atomwndw.cpp
//  AUTHOR:       Richard S. Wright Jr.
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TAtomWindow (TWindow).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>
#include "atomapp.h"
#include "atomwndw.h"
#include <stdio.h>

// Forward declarations
void GLResize(GLsizei w, GLsizei h);
void GLRenderScene(void *pData);
void GLSetupRC(void *pData);

// Easier way to specify color for Windows bigots
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)


//
// Build a response table for all messages/commands handled by the application.
//
DEFINE_RESPONSE_TABLE1(TAtomWindow, TWindow)
//{{TAtomWindowRSP_TBL_BEGIN}}
  EV_WM_DESTROY,
  EV_WM_ERASEBKGND,
  EV_WM_CREATE,
  EV_WM_PAINT,
  EV_WM_SIZE,
  EV_WM_QUERYNEWPALETTE,
  EV_WM_PALETTECHANGED,
  EV_WM_TIMER,
//{{TAtomWindowRSP_TBL_END}}
END_RESPONSE_TABLE;


//{{TAtomWindow Implementation}}


//--------------------------------------------------------
// TAtomWindow
// ~~~~~~~~~~
// Construction/Destruction handling.
//
TAtomWindow::TAtomWindow(TWindow* parent, const char far* title, TModule* module)
:
  TWindow(parent, title, module)
{
  	m_pPalette = NULL;
	m_hDC = NULL;
}


TAtomWindow::~TAtomWindow()
	{
	Destroy();
  	delete m_pPalette;
	}

void TAtomWindow::EvDestroy()
{
	// Kill the timer
   KillTimer(101);

   // Free the rendering context
	wglMakeCurrent(NULL,NULL);
   wglDeleteContext(m_hRC);

	// Release the device context
   ::ReleaseDC(this->GetHandle(),m_hDC);

   TWindow::EvDestroy();
   }


bool TAtomWindow::EvEraseBkgnd(HDC dc)
	{
	return FALSE;
	}


int TAtomWindow::EvCreate(CREATESTRUCT far& createStruct)
	{
  int result;

  createStruct.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);

  result = TWindow::EvCreate(createStruct);

  	// Select pixel format/rendering context
	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,					// Version of this structure
		PFD_DRAW_TO_WINDOW |	// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |	// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,		// Double buffered mode
		PFD_TYPE_RGBA,		// RGBA Color mode
		24,				// Want 24bit color
		0,0,0,0,0,0,			// Not used to select mode
		0,0,				// Not used to select mode
		0,0,0,0,0,			// Not used to select mode
		32,				// Size of depth buffer
		0,				// Not used to select mode
		0,				// Not used to select mode
		PFD_MAIN_PLANE,		// Draw in main plane
		0,				// Not used to select mode
		0,0,0 };			// Not used to select mode

	// Get the device context
	m_hDC = ::GetDC(this->GetHandle());

	// Choose a pixel format that best matches that described in pfd
	int nPixelFormat = ChoosePixelFormat(m_hDC, &pfd);

	// Set the pixel format for the device context
	SetPixelFormat(m_hDC, nPixelFormat, &pfd);

	// Create a 3-3-2 palette
   SetupPalette(m_hDC);

	// Create the rendering context
	m_hRC = wglCreateContext(m_hDC);

	// Make the rendering context current and perform initializion.
	wglMakeCurrent(m_hDC,m_hRC);
	GLSetupRC(m_hDC);

   // Set a timer for 200 milliseconds
   SetTimer(200,101,NULL);

   return result;
	}


void TAtomWindow::EvPaint()
{
	// Make the rendering context current, and call OpenGL Rendering code
	wglMakeCurrent(m_hDC,m_hRC);
	GLRenderScene(NULL);
	wglMakeCurrent(NULL,m_hRC);

	// Finally swap buffers since this rendering context is double buffered
	SwapBuffers(m_hDC);

	// Validate the window
	Validate();
}


void TAtomWindow::EvSize(uint sizeType, TSize& size)
{
    TWindow::EvSize(sizeType, size);

	// Make the rendering context current, and call function
	// to make adjustments to OpenGL viewport
	wglMakeCurrent(m_hDC,m_hRC);
	GLResize(size.cx, size.cy);
	wglMakeCurrent(m_hDC,NULL);
}


bool TAtomWindow::EvQueryNewPalette()
{
	bool result;

	// Only if palette was created
	if(m_pPalette != NULL)
		{
		int nRet;

		// Select the palette into the current device context
		if(SelectPalette(m_hDC, m_pPalette->GetHandle(),FALSE) == NULL)
			::MessageBox(NULL,"Cannot select Palette","Error",MB_OK);

		// Map entries from the currently selected palette to
		// the system palette.  The return value is the number
		// of palette entries modified.
		nRet = RealizePalette(m_hDC);

		// Repaint, forces remap of palette in current window
		Invalidate();

		return nRet;
		}

	// Call default function
	result = TWindow::EvQueryNewPalette();
	return result;
}


void TAtomWindow::EvPaletteChanged(THandle hWndPalChg)
{
	// Only if palette created, or not this window
	if((m_pPalette != NULL) && (hWndPalChg != this->HWindow))
		{
		// Select the palette into the device context
		::SelectPalette(m_hDC,m_pPalette->GetHandle(),FALSE);

		// Map entries to system palette
		::RealizePalette(m_hDC);

		// Remap the current colors to the newly realized palette
		::UpdateColors(m_hDC);
		return;
		}

	// Call default handler
	TWindow::EvPaletteChanged(hWndPalChg);
}


void TAtomWindow::EvTimer(uint timerId)
{
    TWindow::EvTimer(timerId);

  // Force a repaint
  Invalidate();
}



// Create the palette if necessary
void TAtomWindow::SetupPalette(HDC hDC)
	{
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;					// Counting variable

	BYTE RedRange,GreenRange,BlueRange;  // Range for each color entrys
						    // 			(7,7,and 3)

	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat,
		sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return;

	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;

	// Allocate space for a logical palette structure plus all the
	// palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE)
	+nColors*sizeof(PALETTEENTRY));

	// Fill in palette header
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; 	// table size

	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;

	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
		{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);

		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) &
										GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);

		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);

		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
		}

	// Create the palette
	m_pPalette = new TPalette(pPal);

	// Go ahead and select and realize the palette for this device context
	if(SelectPalette(hDC,m_pPalette->GetHandle(),FALSE) == NULL)
		::MessageBox(NULL,"Cannot select Palette in Palette Creation",
								   "Error",MB_OK);

	if(RealizePalette(hDC) == NULL)
		::MessageBox(NULL,"Cannot realize Palette in Palette Creation",
									"Error",MB_OK);

	// Free the memory used for the logical palette structure
	free(pPal);
	}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Called by Windows when it receives the WM_SIZE message.
// Put any code needed here to recalc the viewing volume and
// viewport info
void GLResize(GLsizei w, GLsizei h)
	{
	GLfloat nRange = 100.0f;

	// Prevent a divide by zero
	if(h == 0)
		h = 1;

   if(w == 0)
   	w = 1;

	// Set Viewport to window dimensions
    glViewport(0, 0, w, h);


	// Reset coordinate system
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// Establish clipping volume (left, right, bottom, top, near, far)
    if (w <= h) 
		glOrtho (-nRange, nRange, nRange*h/w, -nRange*h/w, -nRange*2.0f, nRange*2.0f);
    else 
		glOrtho (-nRange*w/h, nRange*w/h, nRange, -nRange, -nRange*2.0f, nRange*2.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Do any initialization of the rendering context here, such as
// setting background colors, setting up lighting, or performing
// preliminary calculations.
void GLSetupRC(void *pData)
	{
	// Light values and coordinates
	glEnable(GL_DEPTH_TEST);	// Hidden surface removal
	glFrontFace(GL_CCW);		// Counter clock-wise polygons face out
	glEnable(GL_CULL_FACE);		// Do not calculate inside of jet

	// Black background
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f );
 	}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Render your OpenGL Scene here.
void GLRenderScene(void *pData)
	{
	// Angle of revolution around the nucleus
	static float fElect1 = 0.0f;

	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Reset the modelview matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	// Translate the whole scene out and into view	
	// This is the initial viewing transformation
	glTranslatef(0.0f, 0.0f, -100.0f);	

	// Red Nucleus
	glRGB(255, 0, 0);
	auxSolidSphere(10.0f);

	// Yellow Electrons
	glRGB(255,255,0);

	// First Electron Orbit
	// Save viewing transformation
	glPushMatrix();

	// Rotate by angle of revolution
	glRotatef(fElect1, 0.0f, 1.0f, 0.0f);

	// Translate out from origin to orbit distance
	glTranslatef(90.0f, 0.0f, 0.0f);

	// Draw the electron
	auxSolidSphere(6.0f);

	// Restore the viewing transformation
	glPopMatrix();

	// Second Electron Orbit
	glPushMatrix();
	glRotatef(45.0f, 0.0f, 0.0f, 1.0f);
	glRotatef(fElect1, 0.0f, 1.0f, 0.0f);
	glTranslatef(-70.0f, 0.0f, 0.0f);
	auxSolidSphere(6.0f);
	glPopMatrix();


	// Third Electron Orbit
	glPushMatrix();
	glRotatef(360.0f-45.0f,0.0f, 0.0f, 1.0f);
	glRotatef(fElect1, 0.0f, 1.0f, 0.0f);
	glTranslatef(0.0f, 0.0f, 60.0f);
	auxSolidSphere(6.0f);
	glPopMatrix();


	// Increment the angle of revolution
	fElect1 += 10.0f;
	if(fElect1 > 360.0f)
		fElect1 = 0.0f;

	// Flush drawing commands
	glFlush();
	}



