//----------------------------------------------------------------------------
//  Project RGBTrngl
//  Waite Group Press
//  Copyright  1996. All Rights Reserved.
//
//  SUBSYSTEM:    RGBTrngl Application
//  FILE:         rgbtrngw.cpp
//  AUTHOR:       Richard S. Wright Jr.
//
//  OVERVIEW
//  ~~~~~~~~
//  Source file for implementation of TRGBTrnglWindow (TWindow).
//
//----------------------------------------------------------------------------

#include <owl/pch.h>
#include "rgbtrnga.h"
#include "rgbtrngw.h"
#include <stdio.h>

// Forward declarations
void GLResize(GLsizei w, GLsizei h);
void GLRenderScene(void *pData);
void GLSetupRC(void *pData);

// Easier way to specify color for Windows bigots
#define glRGB(x, y, z)	glColor3ub((GLubyte)x, (GLubyte)y, (GLubyte)z)


//
// Build a response table for all messages/commands handled by the application.
//
DEFINE_RESPONSE_TABLE1(TRGBTrnglWindow, TWindow)
//{{TRGBTrnglWindowRSP_TBL_BEGIN}}
  EV_WM_CREATE,
  EV_WM_DESTROY,
  EV_WM_ERASEBKGND,
  EV_WM_PAINT,
  EV_WM_SIZE,
  EV_WM_PALETTECHANGED,
  EV_WM_QUERYNEWPALETTE,
//{{TRGBTrnglWindowRSP_TBL_END}}
END_RESPONSE_TABLE;


//{{TRGBTrnglWindow Implementation}}


//--------------------------------------------------------
// TRGBTrnglWindow
// ~~~~~~~~~~
// Construction/Destruction handling.
//
TRGBTrnglWindow::TRGBTrnglWindow(TWindow* parent, const char far* title, TModule* module)
:
  TWindow(parent, title, module)
{
  	m_pPalette = NULL;
	m_hDC = NULL;
}


TRGBTrnglWindow::~TRGBTrnglWindow()
{
	Destroy();
  	delete m_pPalette;
}

int TRGBTrnglWindow::EvCreate(CREATESTRUCT far& createStruct)
{
  int result;

  createStruct.style |= (WS_CLIPCHILDREN | WS_CLIPSIBLINGS | CS_OWNDC);

  result = TWindow::EvCreate(createStruct);

  	// Select pixel format/rendering context
	static PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),	// Size of this structure
		1,					// Version of this structure
		PFD_DRAW_TO_WINDOW |	// Draw to Window (not to bitmap)
		PFD_SUPPORT_OPENGL |	// Support OpenGL calls in window
		PFD_DOUBLEBUFFER,		// Double buffered mode
		PFD_TYPE_RGBA,		// RGBA Color mode
		24,				// Want 24bit color
		0,0,0,0,0,0,			// Not used to select mode
		0,0,				// Not used to select mode
		0,0,0,0,0,			// Not used to select mode
		32,				// Size of depth buffer
		0,				// Not used to select mode
		0,				// Not used to select mode
		PFD_MAIN_PLANE,		// Draw in main plane
		0,				// Not used to select mode
		0,0,0 };			// Not used to select mode

	// Get the device context
	m_hDC = ::GetDC(this->GetHandle());

	// Choose a pixel format that best matches that described in pfd
	int nPixelFormat = ChoosePixelFormat(m_hDC, &pfd);

	// Set the pixel format for the device context
	SetPixelFormat(m_hDC, nPixelFormat, &pfd);

	// Create a 3-3-2 palette
   SetupPalette(m_hDC);

	// Create the rendering context
	m_hRC = wglCreateContext(m_hDC);

	// Make the rendering context current and perform initializion.
	wglMakeCurrent(m_hDC,m_hRC);
	GLSetupRC(m_hDC);

   return result;
}


void TRGBTrnglWindow::EvDestroy()
{
    TWindow::EvDestroy();

   // Free the rendering context
	wglMakeCurrent(NULL,NULL);
   wglDeleteContext(m_hRC);

	// Release the device context
   ::ReleaseDC(this->GetHandle(),m_hDC);
}


bool TRGBTrnglWindow::EvEraseBkgnd(HDC dc)
	{
	return FALSE;
   }


void TRGBTrnglWindow::EvPaint()
{
	// Make the rendering context current, and call OpenGL Rendering code
	wglMakeCurrent(m_hDC,m_hRC);
	GLRenderScene(NULL);
	wglMakeCurrent(NULL,m_hRC);

	// Finally swap buffers since this rendering context is double buffered
	SwapBuffers(m_hDC);

	// Validate the window
	Validate();
}


void TRGBTrnglWindow::EvSize(uint sizeType, TSize& size)
{
    TWindow::EvSize(sizeType, size);

	// Make the rendering context current, and call function
	// to make adjustments to OpenGL viewport
	wglMakeCurrent(m_hDC,m_hRC);
	GLResize(size.cx, size.cy);
	wglMakeCurrent(m_hDC,NULL);
}


void TRGBTrnglWindow::EvPaletteChanged(THandle hWndPalChg)
{
	// Only if palette created, or not this window
	if((m_pPalette != NULL) && (hWndPalChg != this->HWindow))
		{
		// Select the palette into the device context
		::SelectPalette(m_hDC,m_pPalette->GetHandle(),FALSE);

		// Map entries to system palette
		::RealizePalette(m_hDC);

		// Remap the current colors to the newly realized palette
		::UpdateColors(m_hDC);
		return;
		}

	// Call default handler
	TWindow::EvPaletteChanged(hWndPalChg);
}


bool TRGBTrnglWindow::EvQueryNewPalette()
{
	bool result;

	// Only if palette was created
	if(m_pPalette != NULL)
		{
		int nRet;

		// Select the palette into the current device context
		if(SelectPalette(m_hDC, m_pPalette->GetHandle(),FALSE) == NULL)
			::MessageBox(NULL,"Cannot select Palette","Error",MB_OK);

		// Map entries from the currently selected palette to
		// the system palette.  The return value is the number
		// of palette entries modified.
		nRet = RealizePalette(m_hDC);

		// Repaint, forces remap of palette in current window
		Invalidate();

		return nRet;
		}

	// Call default function
	result = TWindow::EvQueryNewPalette();
	return result;
}

// Create the palette if necessary
void TRGBTrnglWindow::SetupPalette(HDC hDC)
	{
	PIXELFORMATDESCRIPTOR pfd;	// Pixel Format Descriptor
	LOGPALETTE *pPal;			// Pointer to memory for logical palette
	int nPixelFormat;			// Pixel format index
	int nColors;				// Number of entries in palette
	int i;					// Counting variable

	BYTE RedRange,GreenRange,BlueRange;  // Range for each color entrys
						    // 			(7,7,and 3)

	// Get the pixel format index and retrieve the pixel format description
	nPixelFormat = GetPixelFormat(hDC);
	DescribePixelFormat(hDC, nPixelFormat,
		sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	// Does this pixel format require a palette?  If not, do not create a
	// palette and just return
	if(!(pfd.dwFlags & PFD_NEED_PALETTE))
		return;

	// Number of entries in palette.  8 bits yeilds 256 entries
	nColors = 1 << pfd.cColorBits;

	// Allocate space for a logical palette structure plus all the
	// palette entries
	pPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE)
	+nColors*sizeof(PALETTEENTRY));

	// Fill in palette header
	pPal->palVersion = 0x300;		// Windows 3.0
	pPal->palNumEntries = nColors; 	// table size

	// Build mask of all 1's.  This creates a number represented by having
	// the low order x bits set, where x = pfd.cRedBits, pfd.cGreenBits, and
	// pfd.cBlueBits.
	RedRange = (1 << pfd.cRedBits) -1;
	GreenRange = (1 << pfd.cGreenBits) - 1;
	BlueRange = (1 << pfd.cBlueBits) -1;

	// Loop through all the palette entries
	for(i = 0; i < nColors; i++)
		{
		// Fill in the 8-bit equivalents for each component
		pPal->palPalEntry[i].peRed = (i >> pfd.cRedShift) & RedRange;
		pPal->palPalEntry[i].peRed = (unsigned char)(
			(double) pPal->palPalEntry[i].peRed * 255.0 / RedRange);

		pPal->palPalEntry[i].peGreen = (i >> pfd.cGreenShift) &
										GreenRange;
		pPal->palPalEntry[i].peGreen = (unsigned char)(
			(double)pPal->palPalEntry[i].peGreen * 255.0 / GreenRange);

		pPal->palPalEntry[i].peBlue = (i >> pfd.cBlueShift) & BlueRange;
		pPal->palPalEntry[i].peBlue = (unsigned char)(
			(double)pPal->palPalEntry[i].peBlue * 255.0 / BlueRange);

		pPal->palPalEntry[i].peFlags = (unsigned char) NULL;
		}

	// Create the palette
	m_pPalette = new TPalette(pPal);

	// Go ahead and select and realize the palette for this device context
	if(SelectPalette(hDC,m_pPalette->GetHandle(),FALSE) == NULL)
		::MessageBox(NULL,"Cannot select Palette in Palette Creation",
								   "Error",MB_OK);

	if(RealizePalette(hDC) == NULL)
		::MessageBox(NULL,"Cannot realize Palette in Palette Creation",
									"Error",MB_OK);

	// Free the memory used for the logical palette structure
	free(pPal);
	}

   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Called by Windows when it receives the WM_SIZE message.
// Put any code needed here to recalc the viewing volume and
// viewport info
void GLResize(GLsizei w, GLsizei h)
	{
	GLfloat windowHeight,windowWidth;

	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	if(w == 0)
   	w = 1;

	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	// Reset the coordinate system before modifying
    glLoadIdentity();


	// Keep the square square.

	// Window is higher than wide
	if (w <= h)
		{
		windowHeight = 250.0f*h/w;
		windowWidth = 250.0f;
		}
    else
		{
		// Window is wider than high
		windowWidth = 250.0f*w/h;
		windowHeight = 250.0f;
		}

	// Set the clipping volume
	glOrtho(-windowWidth, windowWidth, -windowHeight, windowHeight, 1.0f, -1.0f);
	}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Do any initialization of the rendering context here, such as
// setting background colors, setting up lighting, or performing
// preliminary calculations.
void GLSetupRC(void *pData)
	{
	// Enable smooth shading
	glShadeModel(GL_SMOOTH);
 	}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
// Render your OpenGL Scene here.
void GLRenderScene(void *pData)
	{
	// Clear the window with current clearing color
	glClear(GL_COLOR_BUFFER_BIT);


	// Draw the triangle
	glBegin(GL_TRIANGLES);
		// Red Apex
		glColor3ub((GLubyte)255,(GLubyte)0,(GLubyte)0);
		glVertex3f(0.0f,200.0f,0.0f);

		// Green on the right bottom corner
		glColor3ub((GLubyte)0,(GLubyte)255,(GLubyte)0);
		glVertex3f(200.0f,-70.0f,0.0f);

		// Blue on the left bottom corner
		glColor3ub((GLubyte)0,(GLubyte)0,(GLubyte)255);
		glVertex3f(-200.0f, -70.0f, 0.0f);
	glEnd();


	// Flush drawing commands
	glFlush();
	}


